# Copyright contributors to the TSFM project
#
"""Tools for building torch datasets"""

import copy
import enum
import logging
from itertools import starmap
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch

from .util import check_nested_lengths, is_nested_dataframe, join_list_without_repeat


LOGGER = logging.getLogger(__file__)


class ImputeMethod(enum.Enum):
    """`Enum` for the different imputation methods ."""

    FORWARD_FILL = "forward_fill"
    LINEAR = "linear"


class BaseDFDataset(torch.utils.data.Dataset):
    """Base dataset for time series models built upon a pandas dataframe

    Args:
        data_df (pd.DataFrame): Underlying pandas dataframe.
        id_columns (List[str], optional): List of columns which contain id information to separate distinct time series. Defaults to [].
        timestamp_column (Optional[str], optional): Name of the timestamp column. Defaults to None.
        group_id (Optional[Union[List[int], List[str]]], optional): _description_. Defaults to None.
        x_cols (list, optional): Columns to treat as inputs. If an empty list ([]) all the columns in the data_df are taken, except the timestamp column. Defaults to [].
        y_cols (list, optional): Columns to treat as outputs. Defaults to [].
        drop_cols (list, optional): List of columns that are dropped to form the X matrix (input). Defaults to [].
        context_length (int, optional): Length of historical data used when creating individual examples in the torch dataset. Defaults to 1.
        prediction_length (int, optional): Length of prediction (future values). Defaults to 0.
        zero_padding (bool, optional): If True, windows of context_length+prediction_length which are too short are padded with zeros. Defaults to True.
        stride (int, optional): Stride at which windows are produced. Defaults to 1.
        fill_value (Union[float, int], optional): Value used to fill any missing values. Defaults to 0.0.
    """

    def __init__(
        self,
        data_df: pd.DataFrame,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        group_id: Optional[Union[List[int], List[str]]] = None,
        x_cols: list = [],
        y_cols: list = [],
        drop_cols: list = [],
        context_length: int = 1,
        prediction_length: int = 0,
        zero_padding: bool = True,
        stride: int = 1,
        fill_value: Union[float, int] = 0.0,
        full_series=False,
    ):
        super().__init__()
        if not isinstance(x_cols, list):
            x_cols = [x_cols]
        if not isinstance(y_cols, list):
            y_cols = [y_cols]

        if len(x_cols) > 0:
            there, missing = is_cols_in_df(data_df, x_cols)
            assert there, f"{missing} given in {x_cols} is not a valid column identifier in the data."

        if len(y_cols) > 0:
            there, missing = is_cols_in_df(data_df, y_cols)
            assert there, f"{missing} given in {y_cols} is not a valid column identifier in the data."

        if timestamp_column:
            assert timestamp_column in list(
                data_df.columns
            ), f"{timestamp_column} is not in the list of data column names provided {data_df.columns}"
            assert (
                timestamp_column not in x_cols
            ), f"{timestamp_column} can not be used as a timestamp column as it also appears in provided collection:{x_cols}."

        self.data_df = data_df
        self.datetime_col = timestamp_column
        self.id_columns = id_columns
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.drop_cols = drop_cols
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.zero_padding = zero_padding
        self.fill_value = fill_value
        self.timestamps = None
        self.group_id = group_id
        self.stride = stride

        # sort the data by datetime
        if timestamp_column in list(data_df.columns):
            # if not isinstance(data_df[timestamp_column].iloc[0], pd.Timestamp):
            #     data_df[timestamp_column] = pd.to_datetime(data_df[timestamp_column])
            data_df = data_df.sort_values(timestamp_column, ignore_index=True)

        # pad zero to the data_df if the len is shorter than seq_len+pred_len
        # currently no padding when in full_series mode
        if not full_series:
            if zero_padding:
                data_df = self.pad_zero(data_df)
            elif len(data_df) < self.context_length + self.prediction_length:
                LOGGER.warning(
                    f"Padding is disabled and input data is shorter than required length. Received {len(data_df)} time point, but require at least {self.context_length + self.prediction_length} time points."
                )

        if timestamp_column in list(data_df.columns):
            self.timestamps = data_df[timestamp_column].to_list()  # .values coerces timestamps
        # get the input data
        if len(x_cols) > 0:
            self.X = data_df[x_cols]
        else:
            drop_cols = self.drop_cols + y_cols
            if timestamp_column:
                drop_cols += [timestamp_column]
            self.X = data_df.drop(drop_cols, axis=1) if len(drop_cols) > 0 else data_df
            self.x_cols = list(self.X.columns)

        # get target data
        if len(y_cols) > 0:
            self.y = data_df[y_cols]
        else:
            self.y = None

        # get number of X variables
        self.n_vars = self.X.shape[1]
        # get number of target
        self.n_targets = len(y_cols) if len(y_cols) > 0 else 0

    def pad_zero(self, data_df):
        # return zero_padding_to_df(data_df, self.seq_len + self.pred_len)
        return ts_padding(
            data_df,
            timestamp_column=self.datetime_col,
            id_columns=self.id_columns,
            context_length=self.context_length + self.prediction_length,
        )

    def __len__(self):
        return max((len(self.X) - self.context_length - self.prediction_length) // self.stride + 1, 0)

    def _check_index(self, index: int) -> int:
        if index >= len(self):
            raise IndexError("Index exceeds dataset length")

        if index < 0:
            if -index > len(self):
                raise ValueError("Absolute value of index should not exceed dataset length")
            index = len(self) + index
        return index

    def __getitem__(self, index: int):
        """
        Args:
            index (int): Index
        Returns:
            (Any): Sample and meta data, optionally transformed by the respective transforms.
        """
        raise NotImplementedError


class BaseConcatDFDataset(torch.utils.data.ConcatDataset):
    """A dataset consisting of a concatenation of other datasets, based on torch ConcatDataset.

    Args:
        data_df (pd.DataFrame): Underlying pandas dataframe.
        id_columns (List[str], optional): List of columns which contain id information to separate distinct time series. Defaults to [].
        timestamp_column (Optional[str], optional): Name of the timestamp column. Defaults to None.
        context_length (int, optional): Length of historical data used when creating individual examples in the torch dataset. Defaults to 1.
        prediction_length (int, optional): Length of prediction (future values). Defaults to 0.
        num_workers (int, optional): (Currently not used) Number of workers. Defaults to 1.
        fill_value (Union[float, int], optional): Value used to fill any missing values. Defaults to 0.0.
        cls (_type_, optional): The dataset class used to create the underlying datasets. Defaults to BaseDFDataset.
        stride (int, optional): Stride at which windows are produced. Defaults to 1.
    """

    def __init__(
        self,
        data_df: pd.DataFrame,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        context_length: int = 1,
        prediction_length: int = 1,
        num_workers: int = 1,
        fill_value: Union[float, int] = 0.0,
        cls=BaseDFDataset,
        stride: int = 1,
        **kwargs,
    ):
        if len(id_columns) > 0:
            there, missing = is_cols_in_df(data_df, id_columns)
            assert there, f"{missing} given in {id_columns} is not a valid column in the data."

        self.timestamp_column = timestamp_column
        self.id_columns = id_columns
        # self.x_cols = x_cols
        # self.y_cols = y_cols
        self.context_length = context_length
        self.num_workers = num_workers
        self.cls = cls
        self.prediction_length = prediction_length
        self.stride = stride
        self.extra_kwargs = kwargs
        self.fill_value = fill_value
        self.cls = cls

        # create groupby object
        if len(id_columns) == 1:
            self.group_df = data_df.groupby(by=self.id_columns[0])
        elif len(id_columns) > 1:
            self.group_df = data_df.groupby(by=self.id_columns)
        else:
            data_df["group"] = 0  # create a artificial group
            self.group_df = data_df.groupby(by="group")

        # add group_ids to the drop_cols
        self.drop_cols = id_columns if len(id_columns) > 0 else ["group"]

        self.group_names = list(self.group_df.groups.keys())
        datasets = self.concat_dataset()
        super().__init__(datasets)
        self.n_vars = self.datasets[0].n_vars
        self.n_targets = self.datasets[0].n_targets

    def concat_dataset(self):
        """Create a list of Datasets

        Returns:
            List of datasets
        """
        group_df = self.group_df
        # print(f'group_df: {group_df}')
        # pool = mp.Pool(self.num_workers)
        # pool.starmap(
        list_dset = starmap(
            get_group_data,
            [
                (
                    self.cls,
                    group,
                    group_id,
                    self.id_columns,
                    self.timestamp_column,
                    self.context_length,
                    self.prediction_length,
                    self.drop_cols,
                    self.stride,
                    self.fill_value,
                    self.extra_kwargs,
                )
                for group_id, group in group_df
            ],
        )

        # pool.close()
        # del group_df
        return list_dset


def get_group_data(
    cls,
    group,
    group_id,
    id_columns: List[str] = [],
    timestamp_column: Optional[str] = None,
    context_length: int = 1,
    prediction_length: int = 1,
    drop_cols: Optional[List[str]] = None,
    stride: int = 1,
    fill_value: Union[float, int] = 0.0,
    extra_kwargs: Dict[str, Any] = {},
):
    return cls(
        data_df=group,
        group_id=group_id if isinstance(group_id, tuple) else (group_id,),
        id_columns=id_columns,
        timestamp_column=timestamp_column,
        context_length=context_length,
        prediction_length=prediction_length,
        drop_cols=drop_cols,
        stride=stride,
        fill_value=fill_value,
        **extra_kwargs,
    )


class PretrainDFDataset(BaseConcatDFDataset):
    """A dataset used for masked pre-training.

    Args:
        data (pd.DataFrame): Underlying pandas dataframe.
        id_columns (List[str], optional): List of columns which contain id information to separate distinct time series. Defaults to [].
        timestamp_column (Optional[str], optional): Name of the timestamp column. Defaults to None.
        target_columns (List[str], optional): List of column names which identify the target channels in the input, these are the
            columns that will be predicted. Defaults to [].
        context_length (int, optional): Length of historical data used when creating individual examples in the torch dataset. Defaults to 1.
        num_workers (int, optional): (Currently not used) Number of workers. Defaults to 1.
        stride (int, optional): Stride at which windows are produced. Defaults to 1.
        fill_value (Union[float, int], optional): Value used to fill any missing values. Defaults to 0.0.
        enable_padding (bool, optional): If True, windows of context_length+prediction_length which are too short are padded with zeros. Defaults to True.
            If False, a warning is issued when the input data does not contain sufficient records to create a non-empty dataset.



    The resulting dataset returns records (dictionaries) containing:
        past_values: tensor of past values of the target columns of length equal to context length
        past_observed_mask: tensor indicating which values are observed in the past values tensor
        timestamp: the timestamp of the end of the context window
        id: a tuple of id values (taken from the id columns) containing the id information of the time series segment
    """

    def __init__(
        self,
        data: pd.DataFrame,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        target_columns: List[str] = [],
        context_length: int = 1,
        num_workers: int = 1,
        stride: int = 1,
        fill_value: Union[float, int] = 0.0,
        enable_padding: bool = True,
    ):
        super().__init__(
            data_df=data,
            id_columns=id_columns,
            timestamp_column=timestamp_column,
            num_workers=num_workers,
            context_length=context_length,
            prediction_length=0,
            cls=self.BasePretrainDFDataset,
            target_columns=target_columns,
            stride=stride,
            fill_value=fill_value,
            enable_padding=enable_padding,
        )
        self.n_inp = 1

    class BasePretrainDFDataset(BaseDFDataset):
        def __init__(
            self,
            data_df: pd.DataFrame,
            group_id: Optional[Union[List[int], List[str]]] = None,
            context_length: int = 1,
            prediction_length: int = 0,
            drop_cols: list = [],
            id_columns: List[str] = [],
            timestamp_column: Optional[str] = None,
            target_columns: List[str] = [],
            stride: int = 1,
            fill_value: Union[float, int] = 0.0,
            enable_padding: bool = True,
        ):
            self.target_columns = target_columns

            x_cols = target_columns
            y_cols = []

            super().__init__(
                data_df=data_df,
                id_columns=id_columns,
                timestamp_column=timestamp_column,
                x_cols=x_cols,
                y_cols=y_cols,
                context_length=context_length,
                prediction_length=prediction_length,
                group_id=group_id,
                drop_cols=drop_cols,
                stride=stride,
                fill_value=fill_value,
                zero_padding=enable_padding,
            )

        def __getitem__(self, index):
            index = self._check_index(index)

            time_id = index * self.stride
            seq_x = self.X.iloc[time_id : time_id + self.context_length].values
            ret = {
                "past_values": np_to_torch(np.nan_to_num(seq_x, nan=self.fill_value)),
                "past_observed_mask": np_to_torch(~np.isnan(seq_x)),
            }
            if self.datetime_col:
                ret["timestamp"] = self.timestamps[time_id + self.context_length - 1]
            if self.group_id:
                ret["id"] = self.group_id

            return ret


class ForecastDFDataset(BaseConcatDFDataset):
    """A dataset used for forecasting pretraing and inference

    Args:
        data (pd.DataFrame): Underlying pandas dataframe.
        id_columns (List[str], optional): List of columns which contain id information to separate distinct time series. Defaults
            to [].
        timestamp_column (Optional[str], optional): Name of the timestamp column. Defaults to None.
        target_columns (List[str], optional): List of column names which identify the target channels in the input, these are the
            columns that will be predicted. Defaults to [].
        observable_columns (List[str], optional): List of column names which identify the observable channels in the input.
            Observable channels are channels which we have knowledge about in the past and future. For example, weather
            conditions such as temperature or precipitation may be known or estimated in the future, but cannot be
            changed. Defaults to [].
        control_columns (List[str], optional): List of column names which identify the control channels in the input. Control
            channels are similar to observable channels, except that future values may be controlled. For example, discount
            percentage of a particular product is known and controllable in the future. Defaults to [].
        conditional_columns (List[str], optional): List of column names which identify the conditional channels in the input.
            Conditional channels are channels which we know in the past, but do not know in the future. Defaults to [].
        categorical_columns (List[str]): List of column names which identify time-varying categorical-valued channels in the input.
            Defaults to [].
        static_categorical_columns (List[str], optional): List of column names which identify categorical-valued channels in the
            input which are fixed over time. Defaults to [].
        context_length (int, optional): Length of historical data used when creating individual examples in the torch dataset.
            Defaults to 1.
        prediction_length (int, optional): Length of the future forecast. Defaults to 1.
        num_workers (int, optional): (Currently not used) Number of workers. Defaults to 1.
        frequency_token (Optional[int], optional): An integer representing the frequency of the data. Please see for an example of
            frequency token mappings. Defaults to None.
        autoregressive_modeling (bool, optional): (Experimental) If False, any target values in the context window are masked and
            replaced by 0. If True, the context window contains all the historical target information. Defaults to True.
        stride (int, optional): Stride at which windows are produced. Defaults to 1.
        fill_value (Union[float, int], optional): Value used to fill any missing values. Defaults to 0.0.
        masking_specification (List[Tuple[str, Union[int, Tuple[int, int]]]], optional): Allow masking the history (past values) of
            specific columns. The complete specification is a list of individual column masking specifications. A column masking
            specification is a 2-tuple where the first index specifies a column name. The second index specifies and index/indices to
            mask in each of the the context windows (past_values) generated for that column. If a single index is provided, masking
            will begin at that index and continue to the end of the context window. If a tuple of two values is provded, they are
            treated as python list indices; the values given by these indices will be masked.
        enable_padding (bool, optional): If True, windows of context_length+prediction_length which are too short are padded with zeros. Defaults to True.
            If False, a warning is issued when the input data does not contain sufficient records to create a non-empty dataset.
        metadata_columns (List[str], optional): List of columns to be passed through to the output of the dataset as `metadata`.
            Defaults to [].
        impute_method (str, optional): Enables imputation in the past_values of the dataset output. Possible options are in the
            ImputeMethod enum, "forward_fill" fills the values using previous values. Any values which cannot be filled are filled with
            `fill_value`. "linear" use linear interpolation to imput the missing values. None: use regular filling with `fill_value`.
            Defaults to None.

    The resulting dataset returns records (dictionaries) containing:
        past_values: tensor of past values of the target columns of length equal to context length (context_length x number of features)
        past_observed_mask: tensor indicating which values are observed in the past values tensor (context_length x number of features)
        future_values: tensor of future values of the target columns of length equal to prediction length (prediction_length x number of features)
        future_observed_mask: tensor indicating which values are observed in the future values tensor (prediction_length x number of features)
        freq_token: tensor containing the frequency token (scalar)
        static_categorical_features: tensor of static categorical features (1 x len(static_categorical_columns))
        timestamp: the timestamp of the end of the context window
        id: a tuple of id values (taken from the id columns) containing the id information of the time series segment

        where number of features is the total number of columns specified in target_columns, observable_columns, control_columns,
            conditional_columns
    """

    def __init__(
        self,
        data: pd.DataFrame,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        target_columns: List[str] = [],
        observable_columns: List[str] = [],
        control_columns: List[str] = [],
        conditional_columns: List[str] = [],
        categorical_columns: List[str] = [],
        static_categorical_columns: List[str] = [],
        context_length: int = 1,
        prediction_length: int = 1,
        num_workers: int = 1,
        frequency_token: Optional[int] = None,
        autoregressive_modeling: bool = True,
        stride: int = 1,
        fill_value: Union[float, int] = 0.0,
        masking_specification: Optional[List[Tuple[str, Union[int, Tuple[int, int]]]]] = None,
        enable_padding: bool = True,
        metadata_columns: List[str] = [],
        impute_method: Optional[str] = None,
    ):
        # output_columns_tmp = input_columns if output_columns == [] else output_columns

        if not (
            impute_method is None
            or impute_method == ImputeMethod.FORWARD_FILL.value
            or impute_method == ImputeMethod.LINEAR.value
        ):
            raise ValueError(f"Unknown impute_method {self.impute_method}.")

        super().__init__(
            data_df=data,
            id_columns=id_columns,
            timestamp_column=timestamp_column,
            num_workers=num_workers,
            context_length=context_length,
            prediction_length=prediction_length,
            fill_value=fill_value,
            cls=self.BaseForecastDFDataset,
            stride=stride,
            enable_padding=enable_padding,
            # extra_args
            target_columns=target_columns,
            observable_columns=observable_columns,
            control_columns=control_columns,
            conditional_columns=conditional_columns,
            categorical_columns=categorical_columns,
            static_categorical_columns=static_categorical_columns,
            frequency_token=frequency_token,
            autoregressive_modeling=autoregressive_modeling,
            masking_specification=masking_specification,
            metadata_columns=metadata_columns,
            impute_method=impute_method,
        )
        self.n_inp = 2
        # for forecasting, the number of targets is the same as number of X variables
        self.n_targets = self.n_vars

    class BaseForecastDFDataset(BaseDFDataset):
        """
        X_{t+1,..., t+p} = f(X_{:t})
        """

        def __init__(
            self,
            data_df: pd.DataFrame,
            group_id: Optional[Union[List[int], List[str]]] = None,
            context_length: int = 1,
            prediction_length: int = 1,
            drop_cols: list = [],
            id_columns: List[str] = [],
            timestamp_column: Optional[str] = None,
            target_columns: List[str] = [],
            observable_columns: List[str] = [],
            control_columns: List[str] = [],
            conditional_columns: List[str] = [],
            categorical_columns: List[str] = [],
            static_categorical_columns: List[str] = [],
            frequency_token: Optional[int] = None,
            autoregressive_modeling: bool = True,
            stride: int = 1,
            fill_value: Union[float, int] = 0.0,
            masking_specification: Optional[List[Tuple[str, Union[int, Tuple[int, int]]]]] = None,
            enable_padding: bool = True,
            metadata_columns: List[str] = [],
            impute_method: Optional[str] = ImputeMethod.FORWARD_FILL.value,
        ):
            self.frequency_token = frequency_token
            self.target_columns = target_columns
            self.observable_columns = observable_columns
            self.control_columns = control_columns
            self.conditional_columns = conditional_columns
            self.categorical_columns = categorical_columns
            self.static_categorical_columns = static_categorical_columns
            self.autoregressive_modeling = autoregressive_modeling
            self.masking_specification = masking_specification
            self.metadata_columns = metadata_columns
            self.impute_method = impute_method

            x_cols = join_list_without_repeat(
                target_columns,
                observable_columns,
                control_columns,
                conditional_columns,
                categorical_columns,
            )
            y_cols = copy.copy(x_cols)

            self.column_name_to_index_map = {k: v for v, k in enumerate(x_cols)}

            # check non-autoregressive case
            if len(target_columns) == len(x_cols) and not self.autoregressive_modeling:
                raise ValueError(
                    "Non-autoregressive modeling was chosen, but there are no input columns for prediction."
                )

            # masking for conditional values which are not observed during future period
            self.y_mask_conditional = np.array([(c in conditional_columns) for c in y_cols])

            # create a mask of x which masks targets
            self.x_mask_targets = np.array([(c in target_columns) for c in x_cols])

            super().__init__(
                data_df=data_df,
                id_columns=id_columns,
                timestamp_column=timestamp_column,
                x_cols=x_cols,
                y_cols=y_cols,
                context_length=context_length,
                prediction_length=prediction_length,
                group_id=group_id,
                drop_cols=drop_cols,
                stride=stride,
                fill_value=fill_value,
                zero_padding=enable_padding,
            )

        def __getitem__(self, index):
            # seq_x: batch_size x seq_len x num_x_cols
            index = self._check_index(index)

            time_id = index * self.stride

            seq_x = self.X.iloc[time_id : time_id + self.context_length].values.astype(np.float32)
            if not self.autoregressive_modeling:
                seq_x[:, self.x_mask_targets] = 0

            if self.masking_specification is not None:
                seq_x = apply_masking_specification(
                    seq_x,
                    masking_specification=self.masking_specification,
                    column_name_to_index_map=self.column_name_to_index_map,
                )

            if self.impute_method == ImputeMethod.FORWARD_FILL.value:
                seq_x_imputed = impute_forward_fill(seq_x, self.fill_value)
            elif self.impute_method == ImputeMethod.LINEAR.value:
                # interpolate for each channel
                seq_x_imputed = np.apply_along_axis(interpolate_by_var, 0, seq_x)
            elif self.impute_method is None:
                seq_x_imputed = np.nan_to_num(seq_x, nan=self.fill_value)
            else:
                raise ValueError(f"Unknown impute_method {self.impute_method}.")

            # seq_y: batch_size x pred_len x num_x_cols
            seq_y = self.y.iloc[
                time_id + self.context_length : time_id + self.context_length + self.prediction_length
            ].values
            seq_y[:, self.y_mask_conditional] = 0

            ret = {
                "past_values": np_to_torch(seq_x_imputed),
                "future_values": np_to_torch(np.nan_to_num(seq_y, nan=self.fill_value)),
                "past_observed_mask": np_to_torch(~np.isnan(seq_x)),
                "future_observed_mask": np_to_torch(~np.isnan(seq_y)),
            }

            if self.datetime_col:
                ret["timestamp"] = self.timestamps[time_id + self.context_length - 1]

            if self.group_id:
                ret["id"] = self.group_id

            if self.frequency_token is not None:
                ret["freq_token"] = torch.tensor(self.frequency_token, dtype=torch.int)

            if self.static_categorical_columns:
                categorical_values = self.data_df[self.static_categorical_columns].values[0, :]
                ret["static_categorical_values"] = np_to_torch(categorical_values)

            if self.metadata_columns:
                ret["metadata"] = self.data_df[self.metadata_columns].values[
                    time_id : time_id + self.context_length + self.prediction_length, :
                ]

            return ret

        def __len__(self):
            return max((len(self.X) - self.context_length - self.prediction_length) // self.stride + 1, 0)


class ImputeForecastDFDataset(BaseConcatDFDataset):
    """A dataset used for forecasting pretraing and inference

    Args:
        data (pd.DataFrame): Underlying pandas dataframe.
        id_columns (List[str], optional): List of columns which contain id information to separate distinct time series. Defaults
            to [].
        timestamp_column (Optional[str], optional): Name of the timestamp column. Defaults to None.
        target_columns (List[str], optional): List of column names which identify the target channels in the input, these are the
            columns that will be predicted. Defaults to [].
        observable_columns (List[str], optional): List of column names which identify the observable channels in the input.
            Observable channels are channels which we have knowledge about in the past and future. For example, weather
            conditions such as temperature or precipitation may be known or estimated in the future, but cannot be
            changed. Defaults to [].
        control_columns (List[str], optional): List of column names which identify the control channels in the input. Control
            channels are similar to observable channels, except that future values may be controlled. For example, discount
            percentage of a particular product is known and controllable in the future. Defaults to [].
        conditional_columns (List[str], optional): List of column names which identify the conditional channels in the input.
            Conditional channels are channels which we know in the past, but do not know in the future. Defaults to [].
        static_categorical_columns (List[str], optional): List of column names which identify categorical-valued channels in the
            input which are fixed over time. Defaults to [].
        context_length (int, optional): Length of historical data used when creating individual examples in the torch dataset.
            Defaults to 1.
        prediction_length (int, optional): Length of the future forecast. Defaults to 1.
        num_workers (int, optional): (Currently not used) Number of workers. Defaults to 1.
        frequency_token (Optional[int], optional): An integer representing the frequency of the data. Please see for an example of
            frequency token mappings. Defaults to None.
        stride (int, optional): Stride at which windows are produced. Defaults to 1.
        fill_value (Union[float, int], optional): Value used to fill any missing values. Defaults to 0.0.
        masking_specification (List[Tuple[str, Union[int, Tuple[int, int]]]], optional): Allow masking the history (past values) of
            specific columns. The complete specification is a list of individual column masking specifications. A column masking
            specification is a 2-tuple where the first index specifies a column name. The second index specifies and index/indices to
            mask in each of the the context windows (past_values) generated for that column. If a single index is provided, masking
            will begin at that index and continue to the end of the context window. If a tuple of two values is provded, they are
            treated as python list indices; the values given by these indices will be masked.
        artificial_missing_rate (Float, optional): Rate at which points are selected to be marked as artificially missing. Defaults
            to 0.0 (no artificial missing).
        artificial_missing_columns (List[str], optional): Column names on which to apply artificial missing. Defaults to None.
        artificial_missing_at_time_t (bool, optional): If True the point at time t (last point in the context window) is marked as
            missing. Defaults to False.
        metadata_columns (List[str], optional): List of columns to be passed through to the output of the dataset as `metadata`.
            Defaults to [].
        impute_method (str, optional): Enables imputation in the past_values of the dataset output. Possible options are in the
            ImputeMethod enum, "forward_fill" fills the values using previous values. Any values which cannot be filled are filled with
            `fill_value`. "linear" use linear interpolation to imput the missing values. None: use regular filling with `fill_value`.
            Defaults to None.

    The resulting dataset returns records (dictionaries) containing:
        past_values: tensor of past values of the target columns of length equal to context length (context_length x number of features)
        past_observed_mask: tensor indicating which values are observed in the past values tensor (context_length x number of features)
        future_values: tensor of future values of the target columns of length equal to prediction length (prediction_length x number of features)
        future_observed_mask: tensor indicating which values are observed in the future values tensor (prediction_length x number of features)
        freq_token: tensor containing the frequency token (scalar)
        static_categorical_features: tensor of static categorical features (1 x len(static_categorical_columns))
        timestamp: the timestamp of the end of the context window
        id: a tuple of id values (taken from the id columns) containing the id information of the time series segment

        where number of features is the total number of columns specified in target_columns, observable_columns, control_columns,
            conditional_columns
    """

    def __init__(
        self,
        data: pd.DataFrame,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        target_columns: List[str] = [],
        observable_columns: List[str] = [],
        control_columns: List[str] = [],
        conditional_columns: List[str] = [],
        static_categorical_columns: List[str] = [],
        context_length: int = 1,
        prediction_length: int = 1,
        num_workers: int = 1,
        frequency_token: Optional[int] = None,
        stride: int = 1,
        fill_value: Union[float, int] = 0.0,
        masking_specification: Optional[List[Tuple[str, Union[int, Tuple[int, int]]]]] = None,
        artificial_missing_rate: float = 0.0,
        artificial_missing_columns: Optional[List[str]] = None,
        artificial_missing_at_time_t: bool = False,
        random_seed: int = 0,
        impute_method: Optional[str] = ImputeMethod.FORWARD_FILL.value,
    ):
        # output_columns_tmp = input_columns if output_columns == [] else output_columns

        super().__init__(
            data_df=data,
            id_columns=id_columns,
            timestamp_column=timestamp_column,
            num_workers=num_workers,
            context_length=context_length,
            prediction_length=prediction_length,
            fill_value=fill_value,
            cls=self.BaseImputeForecastDFDataset,
            stride=stride,
            # extra_args
            target_columns=target_columns,
            observable_columns=observable_columns,
            control_columns=control_columns,
            conditional_columns=conditional_columns,
            static_categorical_columns=static_categorical_columns,
            frequency_token=frequency_token,
            masking_specification=masking_specification,
            artificial_missing_rate=artificial_missing_rate,
            artificial_missing_columns=artificial_missing_columns,
            artificial_missing_at_time_t=artificial_missing_at_time_t,
            rng=np.random.default_rng(seed=random_seed),
            impute_method=impute_method,
        )
        self.n_inp = 2
        # for forecasting, the number of targets is the same as number of X variables
        self.n_targets = self.n_vars

    class BaseImputeForecastDFDataset(BaseDFDataset):
        """
        X_{t+1,..., t+p} = f(X_{:t})
        """

        def __init__(
            self,
            data_df: pd.DataFrame,
            group_id: Optional[Union[List[int], List[str]]] = None,
            context_length: int = 1,
            prediction_length: int = 1,
            drop_cols: list = [],
            id_columns: List[str] = [],
            timestamp_column: Optional[str] = None,
            target_columns: List[str] = [],
            observable_columns: List[str] = [],
            control_columns: List[str] = [],
            conditional_columns: List[str] = [],
            static_categorical_columns: List[str] = [],
            frequency_token: Optional[int] = None,
            stride: int = 1,
            fill_value: Union[float, int] = 0.0,
            masking_specification: Optional[List[Tuple[str, Union[int, Tuple[int, int]]]]] = None,
            artificial_missing_rate: float = 0.0,
            artificial_missing_columns: Optional[List[str]] = None,
            artificial_missing_at_time_t: bool = False,
            rng=np.random.default_rng(),
            impute_method: Optional[str] = ImputeMethod.FORWARD_FILL.value,
        ):
            self.frequency_token = frequency_token
            self.target_columns = target_columns
            self.observable_columns = observable_columns
            self.control_columns = control_columns
            self.conditional_columns = conditional_columns
            self.static_categorical_columns = static_categorical_columns
            self.masking_specification = masking_specification
            self.artificial_missing_rate = artificial_missing_rate
            self.artificial_missing_at_time_t = artificial_missing_at_time_t
            self.rng = rng
            self.impute_method = impute_method

            if artificial_missing_columns is None:
                self.artificial_missing_columns = copy.copy(target_columns)
            else:
                self.artificial_missing_columns = artificial_missing_columns

            x_cols = join_list_without_repeat(
                target_columns,
                observable_columns,
                control_columns,
                conditional_columns,
            )
            y_cols = copy.copy(x_cols)

            # masking for conditional values which are not observed during future period
            self._y_mask_conditional = np.array([(c in conditional_columns) for c in y_cols])

            self.column_name_to_index_map = {k: v for v, k in enumerate(x_cols)}

            super().__init__(
                data_df=data_df,
                id_columns=id_columns,
                timestamp_column=timestamp_column,
                x_cols=x_cols,
                y_cols=y_cols,
                context_length=context_length,
                prediction_length=prediction_length,
                group_id=group_id,
                drop_cols=drop_cols,
                stride=stride,
                fill_value=fill_value,
            )

            self._X_imputed = self.X.ffill().fillna(self.fill_value)
            # create a artificially_observed_mask
            # target columns only
            # we don't mask what is already missing

            self._artificial_missing_column_mask = np.array([(c in self.artificial_missing_columns) for c in x_cols])
            artificial_observed_mask = np.ones_like(self.X.values[:, self._artificial_missing_column_mask], dtype=bool)

            if artificial_missing_rate > 0:
                # obs_x = np.where(~np.isnan(self.X.values[:, self._artificial_missing_mask]))
                obs_x = np.nonzero(~np.isnan(self.X.values[:, self._artificial_missing_column_mask]))

                # process columnwise
                for col_idx in range(artificial_observed_mask.shape[1]):
                    # get a column
                    obs_x_tmp = obs_x[0][obs_x[1] == col_idx]

                    # select missing, and generate 2-tuple of indices
                    # k = int(artificial_observed_mask.shape[0] * self.artificial_missing_rate)
                    # if k >= len(obs_x_tmp):
                    k = int(len(obs_x_tmp) * self.artificial_missing_rate)

                    artificial_missing_indices = (
                        self.rng.choice(
                            obs_x_tmp,
                            k,
                            replace=False,
                        ),
                        np.array([col_idx] * k, dtype=int),
                    )
                    artificial_observed_mask[artificial_missing_indices] = False

            self._artificial_past_observed_mask = artificial_observed_mask

        def __getitem__(self, index):
            # seq_x: batch_size x seq_len x num_x_cols

            time_id = index * self.stride

            seq_x = self.X[time_id : time_id + self.context_length].values

            if self.masking_specification is not None:
                seq_x = apply_masking_specification(
                    seq_x,
                    masking_specification=self.masking_specification,
                    column_name_to_index_map=self.column_name_to_index_map,
                )

            if self.impute_method == ImputeMethod.FORWARD_FILL.value:
                seq_x_imputed = impute_forward_fill(seq_x, self.fill_value)
            elif self.impute_method == ImputeMethod.LINEAR.value:
                # interpolate for each channel
                seq_x_imputed = np.apply_along_axis(interpolate_by_var, 0, seq_x)
            elif self.impute_method is None:
                seq_x_imputed = np.nan_to_num(seq_x, nan=self.fill_value)

            # mask only on seq_x
            artificial_past_observed_seq = np.ones_like(seq_x, dtype=bool)
            artificial_past_observed_seq[:, self._artificial_missing_column_mask] = (
                self._artificial_past_observed_mask[time_id : time_id + self.context_length]
            )

            # enforce missing at time t
            if self.artificial_missing_at_time_t:
                # get columns where time t value is observed
                col_obs = ~np.isnan(seq_x[-1, self._artificial_missing_column_mask])
                artificial_past_observed_seq[-1, self._artificial_missing_column_mask] = ~col_obs

            # seq_y: batch_size x pred_len x num_x_cols
            seq_y = self.y[
                time_id + self.context_length : time_id + self.context_length + self.prediction_length
            ].values

            seq_y[:, self._y_mask_conditional] = 0

            ret = {
                "past_values": np_to_torch(seq_x_imputed),
                "future_values": np_to_torch(np.nan_to_num(seq_y, nan=self.fill_value)),
                "past_observed_mask": np_to_torch(~np.isnan(seq_x)),
                "future_observed_mask": np_to_torch(~np.isnan(seq_y)),
                "artificial_past_observed_mask": np_to_torch(artificial_past_observed_seq),
            }

            if self.datetime_col:
                ret["timestamp"] = self.timestamps[time_id + self.context_length - 1]

            if self.group_id:
                ret["id"] = self.group_id

            if self.frequency_token is not None:
                ret["freq_token"] = torch.tensor(self.frequency_token, dtype=torch.int)

            if self.static_categorical_columns:
                categorical_values = self.data_df[self.static_categorical_columns].values[0, :]
                ret["static_categorical_values"] = np_to_torch(categorical_values)

            return ret

        def __len__(self):
            return (len(self.X) - self.context_length - self.prediction_length) // self.stride + 1


class RegressionDFDataset(BaseConcatDFDataset):
    """A dataset used for forecasting pretraing and inference

    Args:
        data (pd.DataFrame): Underlying pandas dataframe.
        id_columns (List[str], optional): List of columns which contain id information to separate distinct time series. Defaults
            to [].
        timestamp_column (Optional[str], optional): Name of the timestamp column. Defaults to None.
        input_columns (List[str], optional): List of columns to use as inputs to the regression
        target_columns (List[str], optional): List of column names which identify the target channels in the input, these are the
            columns that will be predicted. Defaults to [].
        context_length (int, optional): Length of historical data used when creating individual examples in the torch dataset.
            Defaults to 1.
        num_workers (int, optional): (Currently not used) Number of workers. Defaults to 1.
        stride (int, optional): Stride at which windows are produced. Defaults to 1.
        fill_value (Union[float, int], optional): Value used to fill any missing values. Defaults to 0.0.
        enable_padding (bool, optional): If True, windows of context_length+prediction_length which are too short are padded with zeros. Defaults to True.
            If False, a warning is issued when the input data does not contain sufficient records to create a non-empty dataset.


    The resulting dataset returns records (dictionaries) containing:
        past_values: tensor of past values of the target columns of length equal to context length (context_length x len(input_columns))
        past_observed_mask: tensor indicating which values are observed in the past values tensor (context_length x len(input_columns))
        target_values: tensor of future values of the target columns of length equal to prediction length (prediction_length x len(target_columns))
        static_categorical_features: tensor of static categorical features (1 x len(static_categorical_columns))
        timestamp: the timestamp of the end of the context window
        id: a tuple of id values (taken from the id columns) containing the id information of the time series segment
    """

    def __init__(
        self,
        data: pd.DataFrame,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        input_columns: List[str] = [],
        target_columns: List[str] = [],
        static_categorical_columns: List[str] = [],
        context_length: int = 1,
        num_workers: int = 1,
        stride: int = 1,
        fill_value: Union[float, int] = 0.0,
        enable_padding: bool = True,
    ):
        # self.y_cols = y_cols

        super().__init__(
            data_df=data,
            id_columns=id_columns,
            timestamp_column=timestamp_column,
            num_workers=num_workers,
            context_length=context_length,
            prediction_length=0,
            cls=self.BaseRegressionDFDataset,
            input_columns=input_columns,
            target_columns=target_columns,
            static_categorical_columns=static_categorical_columns,
            stride=stride,
            fill_value=fill_value,
            enable_padding=enable_padding,
        )

        self.n_inp = 2

    class BaseRegressionDFDataset(BaseDFDataset):
        """
        y_{t} = f(X_{:t})
        """

        def __init__(
            self,
            data_df: pd.DataFrame,
            group_id: Optional[Union[List[int], List[str]]] = None,
            context_length: int = 1,
            prediction_length: int = 0,
            drop_cols: list = [],
            id_columns: List[str] = [],
            timestamp_column: Optional[str] = None,
            target_columns: List[str] = [],
            input_columns: List[str] = [],
            static_categorical_columns: List[str] = [],
            stride: int = 1,
            fill_value: Union[float, int] = 0.0,
            enable_padding: bool = True,
        ):
            self.target_columns = target_columns
            self.input_columns = input_columns
            self.static_categorical_columns = static_categorical_columns

            x_cols = input_columns
            y_cols = target_columns

            super().__init__(
                data_df=data_df,
                id_columns=id_columns,
                timestamp_column=timestamp_column,
                x_cols=x_cols,
                y_cols=y_cols,
                context_length=context_length,
                prediction_length=prediction_length,
                group_id=group_id,
                drop_cols=drop_cols,
                stride=stride,
                fill_value=fill_value,
                zero_padding=enable_padding,
            )

        def __getitem__(self, index):
            # seq_x: batch_size x seq_len x num_x_cols
            index = self._check_index(index)

            time_id = index * self.stride
            seq_x = self.X.iloc[time_id : time_id + self.context_length].values
            seq_y = self.y.iloc[time_id + self.context_length - 1 : time_id + self.context_length].values.ravel()
            # return _torch(seq_x, seq_y)

            ret = {
                "past_values": np_to_torch(np.nan_to_num(seq_x, nan=self.fill_value)),
                "target_values": np_to_torch(np.nan_to_num(seq_y, nan=self.fill_value)),
                "past_observed_mask": np_to_torch(~np.isnan(seq_x)),
            }
            if self.datetime_col:
                ret["timestamp"] = self.timestamps[time_id + self.context_length - 1]

            if self.group_id:
                ret["id"] = self.group_id

            if self.static_categorical_columns:
                categorical_values = self.data_df[self.static_categorical_columns].values[0, :]
                ret["static_categorical_values"] = np_to_torch(categorical_values)

            return ret


class ClassificationDFDataset(BaseConcatDFDataset):
    """A dataset used for forecasting pretraing and inference

    Args:
        data (pd.DataFrame): Underlying pandas dataframe.
        id_columns (List[str], optional): List of columns which contain id information to separate distinct time series. Defaults
            to [].
        timestamp_column (Optional[str], optional): Name of the timestamp column. Defaults to None.
        input_columns (List[str], optional): List of columns to use as inputs to the regression
        label_column (str, optional): List of column names which identify the label of the time series. Defaults to "label".
        context_length (int, optional): Length of historical data used when creating individual examples in the torch dataset.
            Defaults to 1.
        num_workers (int, optional): (Currently not used) Number of workers. Defaults to 1.
        stride (int, optional): Stride at which windows are produced. Defaults to 1.
        fill_value (Union[float, int], optional): Value used to fill any missing values. Defaults to 0.0.
        enable_padding (bool, optional): If True, windows of context_length+prediction_length which are too short are padded with zeros. Defaults to True.
            If False, a warning is issued when the input data does not contain sufficient records to create a non-empty dataset. Only applies
            when full_series==False.
        full_series (bool, optional): If True, each series is used when creating a single entry in the classification dataset. In this case
            the series is interpolated to match the desired context_length. No padding is done. If False, each series is windowed acording to
            the provided context_length. Defaults to False.


    The resulting dataset returns records (dictionaries) containing:
        past_values: tensor of past values of the target columns of length equal to context length (context_length x len(input_columns))
        past_observed_mask: tensor indicating which values are observed in the past values tensor (context_length x len(input_columns))
        target_values: tensor containing the label (scalar)
        static_categorical_features: tensor of static categorical features (1 x len(static_categorical_columns))
        timestamp: the timestamp of the end of the context window
        id: a tuple of id values (taken from the id columns) containing the id information of the time series segment
    """

    def __init__(
        self,
        data: pd.DataFrame,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        input_columns: List[str] = [],
        label_column: str = "label",
        static_categorical_columns: List[str] = [],
        context_length: int = 1,
        num_workers: int = 1,
        stride: int = 1,
        fill_value: Union[float, int] = 0.0,
        enable_padding: bool = True,
        full_series: bool = True,
    ):
        if full_series:
            if not is_nested_dataframe(data, column=input_columns[0]):
                raise ValueError(
                    "The provided data does not appear to contain row entries which contain a full time series."
                )

            check_nested_lengths(data, columns=input_columns)
            if enable_padding:
                enable_padding = False

            super().__init__(
                data_df=data,
                id_columns=id_columns,
                timestamp_column=timestamp_column,
                num_workers=num_workers,
                context_length=context_length,
                prediction_length=0,
                cls=_FullSeriesClassificationDFDataset,
                input_columns=input_columns,
                label_column=label_column,
                static_categorical_columns=static_categorical_columns,
                stride=stride,
                fill_value=fill_value,
                enable_padding=enable_padding,
            )

        else:
            super().__init__(
                data_df=data,
                id_columns=id_columns,
                timestamp_column=timestamp_column,
                num_workers=num_workers,
                context_length=context_length,
                prediction_length=0,
                cls=_WindowedSeriesClassificationDFDataset,
                input_columns=input_columns,
                label_column=label_column,
                static_categorical_columns=static_categorical_columns,
                stride=stride,
                fill_value=fill_value,
                enable_padding=enable_padding,
            )

        self.n_inp = 2
        self.full_series = full_series


def np_to_torch(data: np.array, float_type=np.float32):
    if data.dtype == "float":
        return torch.from_numpy(data.astype(float_type))
    elif data.dtype == "int":
        return torch.from_numpy(data)
    elif data.dtype == "bool":
        return torch.from_numpy(data)
    return torch.from_numpy(data)


def _torch(*nps):
    return tuple(np_to_torch(x) for x in nps)


def zero_padding_to_df(df: pd.DataFrame, seq_len: int) -> pd.DataFrame:
    """
    check if df has length > seq_len.
    If not, then fill in zero
    Args:
        df (_type_): data frame
        seq_len (int): sequence length
    Returns:
        data frame
    """
    if len(df) >= seq_len:
        return df
    fill_len = seq_len - len(df) + 1
    # add zeros dataframe
    zeros_df = pd.DataFrame(np.zeros([fill_len, df.shape[1]]), columns=df.columns)
    # combine the data
    new_df = pd.concat([zeros_df, df])
    return new_df


def ts_padding(
    df: pd.DataFrame,
    id_columns: Optional[List[str]] = None,
    timestamp_column: Optional[str] = None,
    context_length: int = 1,
) -> pd.DataFrame:
    """
    Pad a dataframe, which is aware of time series conventions.

    Check if df has length >= context_length.
    If not, then fill (prepending) while preserving types and properly handling IDs and dates/timestamps. When
    prepending dates, the sampling interval will be estimated, to create proper preceeding dates.

    The assumption is the provided data contains only one id across the provided ID columns, the value will be
    replicated in the prepended rows.

    Args:
        df (_type_): data frame
        id_columns: List of strings representing columns containing ID information.
        timestamp_column: str for column name containing timestamps.
        context_length (int): required length

    Returns:
        Padded data frame
    """
    l = len(df)
    if l >= context_length:
        return df
    fill_length = context_length - l  # why did we previously have + 1 here?

    # create dataframe
    pad_df = pd.DataFrame(np.zeros([fill_length, df.shape[1]]), columns=df.columns)

    for c in df.columns:
        if (id_columns and c in id_columns) or (c == timestamp_column):
            continue
        pad_df[c] = pad_df[c].astype(df.dtypes[c], copy=False)

    if timestamp_column:
        if len(df) < 2:
            pad_df[timestamp_column] = None
        elif (df[timestamp_column].dtype.type == np.datetime64) or (df[timestamp_column].dtype == int):
            last_timestamp = df.iloc[0][timestamp_column]
            period = df.iloc[1][timestamp_column] - df.iloc[0][timestamp_column]
            prepended_timestamps = [last_timestamp + offset * period for offset in range(-fill_length, 0)]
            pad_df[timestamp_column] = prepended_timestamps
        else:
            pad_df[timestamp_column] = None
        # Ensure same type
        pad_df[timestamp_column] = pad_df[timestamp_column].astype(df[timestamp_column].dtype)

    if id_columns:
        id_values = df.iloc[0][id_columns].to_list()
        for id_column_name, id_column_value in zip(id_columns, id_values):
            pad_df[id_column_name] = id_column_value

    # combine the data
    new_df = pd.concat([pad_df, df])
    return new_df


def is_cols_in_df(df: pd.DataFrame, cols: List[str]) -> bool:
    """
    Args:
        df:
        cols:

    Returns:
        bool
    """
    for col in cols:
        if col not in list(df.columns):
            return False, col
    return True, None


def apply_masking_specification(
    past_values_tensor: np.ndarray,
    masking_specification: List[Tuple[str, Union[int, Tuple[int, int]]]],
    column_name_to_index_map: Dict[str, int],
) -> np.ndarray:
    """Apply the desired mask defined by masking_specification.

    Args:
        past_values_tensor (np.ndarray): Tensor of past values, should have shape (context_length, num_channels)
        masking_specification (List[Tuple[str, Union[int, Tuple[int, int]]]], optional): Allow masking the history (past values) of
            specific columns. The complete specification is a list of individual column masking specifications. A column masking
            specification is a 2-tuple where the first index specifies a column name. The second index specifies and index/indices to
            mask in each of the the context windows (past_values) generated for that column. If a single index is provided, masking
            will begin at that index and continue to the end of the context window. If a tuple of two values is provded, they are
            treated as python list indices; the values given by these indices will be masked.
        column_name_to_index_map (Dict[str, int]): A mapping of string (column names) to the indice of that column

    Returns:
        np.ndarry: Tensor with values masked
    """

    past_values_tensor = past_values_tensor.copy()
    for col_name, spec in masking_specification:
        col_idx = column_name_to_index_map[col_name]
        if isinstance(spec, (tuple, list)) and len(spec) == 2:
            past_values_tensor[spec[0] : spec[1], col_idx] = np.nan
        else:
            past_values_tensor[spec:, col_idx] = np.nan
    return past_values_tensor


def interpolate_by_var(miss_seq: np.ndarray) -> np.ndarray:
    """Interpolate each column of the input numpy array using np.interpolate.

    Args:
        miss_seq (np.ndarray): Sequence potentially containing missing values.

    Returns:
        np.ndarray: Imputed sequence.
    """
    nans = np.isnan(miss_seq)
    if nans.sum() == 0:
        return miss_seq

    imputed = np.interp(np.where(nans)[0], np.where(~nans)[0], miss_seq[~nans])
    miss_seq_copy = miss_seq.copy()
    miss_seq_copy[nans] = imputed
    return miss_seq_copy


def impute_forward_fill(miss_seq: np.ndarray, fill_value: float) -> np.ndarray:
    """Impute each column of the input numpy array using forward filling.

    Args:
        miss_seq (np.ndarray): Sequence potentially containing missing values.
        fill_value (float): Value used to impute any values that cannot be filled

    Returns:
        np.ndarray: Imputed sequence.
    """
    # impute the numpy matrix, time is the first dimension
    # based on https://stackoverflow.com/questions/41190852/most-efficient-way-to-forward-fill-nan-values-in-numpy-array
    mask = np.isnan(miss_seq)
    idx = np.where(~mask, np.expand_dims(np.arange(mask.shape[0]), 1), 0)
    np.maximum.accumulate(idx, axis=0, out=idx)
    miss_seq_copy = miss_seq.copy()
    seq_x_imputed = miss_seq_copy[idx, np.arange(idx.shape[1])[None, :]]

    # fill any remaining nan
    seq_x_imputed = np.nan_to_num(seq_x_imputed, nan=fill_value)
    return seq_x_imputed


class _WindowedSeriesClassificationDFDataset(BaseDFDataset):
    """Standard windowed classification dataset.

    Windows of size context_length are extracted from the input time series. The labels are chosen from the label_column
    at the index of the end of the window.
    """

    def __init__(
        self,
        data_df: pd.DataFrame,
        group_id: Optional[Union[List[int], List[str]]] = None,
        context_length: int = 1,
        prediction_length: int = 0,
        drop_cols: list = [],
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        label_column: str = "label",
        input_columns: List[str] = [],
        static_categorical_columns: List[str] = [],
        stride: int = 1,
        fill_value: Union[float, int] = 0.0,
        enable_padding: bool = True,
    ):
        self.label_column = label_column
        self.input_columns = input_columns
        self.static_categorical_columns = static_categorical_columns

        x_cols = input_columns
        y_cols = label_column

        super().__init__(
            data_df=data_df,
            id_columns=id_columns,
            timestamp_column=timestamp_column,
            x_cols=x_cols,
            y_cols=y_cols,
            context_length=context_length,
            prediction_length=prediction_length,
            group_id=group_id,
            drop_cols=drop_cols,
            stride=stride,
            fill_value=fill_value,
            zero_padding=enable_padding,
            full_series=False,
        )

    def __getitem__(self, index):
        # seq_x: batch_size x seq_len x num_x_cols
        index = self._check_index(index)

        time_id = index * self.stride
        seq_x = self.X.iloc[time_id : time_id + self.context_length].values
        # seq_y = self.y[time_id + self.context_length - 1 : time_id + self.context_length].values.ravel()
        seq_y = self.y.iloc[time_id + self.context_length - 1].values[0]

        ret = {
            "past_values": np_to_torch(np.nan_to_num(seq_x, nan=self.fill_value)),
            "target_values": torch.tensor(np.nan_to_num(seq_y, nan=self.fill_value), dtype=torch.int64),
            "past_observed_mask": np_to_torch(~np.isnan(seq_x)),
        }
        if self.datetime_col:
            ret["timestamp"] = self.timestamps[time_id + self.context_length - 1]

        if self.group_id:
            ret["id"] = self.group_id

        if self.static_categorical_columns:
            categorical_values = self.data_df[self.static_categorical_columns].values[0, :]
            ret["static_categorical_values"] = np_to_torch(categorical_values)

        return ret


class _FullSeriesClassificationDFDataset(BaseDFDataset):
    """Full series classification dataset.

    Works with nested input dataframes (i.e., dataframes which contain pd.Series as entries). In this case, each series
    is interpolated to the desired context_length.
    """

    def __init__(
        self,
        data_df: pd.DataFrame,
        group_id: Optional[Union[List[int], List[str]]] = None,
        context_length: int = 1,
        prediction_length: int = 0,
        drop_cols: list = [],
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        label_column: str = "label",
        input_columns: List[str] = [],
        static_categorical_columns: List[str] = [],
        stride: int = 1,
        fill_value: Union[float, int] = 0.0,
        enable_padding: bool = True,
    ):
        self.label_column = label_column
        self.input_columns = input_columns
        self.static_categorical_columns = static_categorical_columns

        x_cols = input_columns
        y_cols = label_column

        super().__init__(
            data_df=data_df,
            id_columns=id_columns,
            timestamp_column=timestamp_column,
            x_cols=x_cols,
            y_cols=y_cols,
            context_length=context_length,
            prediction_length=prediction_length,
            group_id=group_id,
            drop_cols=drop_cols,
            stride=stride,
            fill_value=fill_value,
            zero_padding=enable_padding,
            full_series=True,
        )

        # after init
        # self.X should contain the data including all inputs (i.e. input columns)
        # X is in the nexted format where X has series entries in some places

    def __len__(self):
        return len(self.X) // self.stride

    def __getitem__(self, index):
        index = self._check_index(index)

        ind = index * self.stride

        # each row has entries which are series
        # assume also that static categorical are static for each row (not necessarily across rows)

        # first convert rows of series to full tensor: len x features
        # assumes consistent length
        X = np.asarray([s.to_list() for s in self.X.iloc[ind].values])
        X = np_to_torch(np.nan_to_num(X, nan=self.fill_value))
        y = np_to_torch(self.y.iloc[ind].values)

        X = X.unsqueeze(dim=0)  # transpose(1, 0). c l --->  1 c l
        X = torch.nn.functional.interpolate(
            X, self.context_length, mode="linear"
        )  # need a 3D tensor of shape B C L    # 1 c l ---> 1 c context_points
        X = X.squeeze(dim=0).transpose(0, 1)  # 1 c cp ---> cp c

        y = torch.squeeze(y)

        ret = {
            "past_values": X,
            "target_values": y,
            # omit or we could do: "past_observed_mask": np_to_torch(np.ones(X.shape)),
        }

        if self.datetime_col:
            if isinstance(self.timestamps[ind], pd.Series):
                ret["timestamp"] = self.timestamps[ind].iloc[-1]
            else:
                ret["timestamp"] = self.timestamps[ind]

        # concatenate the row to the id
        ret["id"] = self.group_id + (ind,) if self.group_id else (ind,)

        if self.static_categorical_columns:
            categorical_values = self.data_df[self.static_categorical_columns].iloc[ind].values
            ret["static_categorical_values"] = np_to_torch(categorical_values)

        return ret


if __name__ == "__main__":
    df = pd.DataFrame(
        {
            "A": [1, 2, 3, 4, 5, 6, 7, 8],
            "B": [4, 5, 6, 7, 8, 9, 10, 11],
            "C": [7, 8, 9, 10, 11, 12, 13, 14],
            "g1": [0, 1, 1, 1, 0, 0, 0, 0],
        }
    )
    print(df)

    d6 = PretrainDFDataset(data_df=df, x_cols=["A", "B"], group_ids=["g1"], seq_len=2)
    print(f"d6: {d6}")

    d7 = ForecastDFDataset(data_df=df, x_cols=["A", "B"], group_ids=["g1"], seq_len=2, pred_len=2)



# Copyright contributors to the TSFM project
#

"""Tests recursive prediction functions"""

import tempfile

from transformers import Trainer, TrainingArguments

from tsfm_public.toolkit.dataset import (
    ForecastDFDataset,
)
from tsfm_public.toolkit.recursive_predictor import RecursivePredictor, RecursivePredictorConfig
from tsfm_public.toolkit.util import select_by_index


def get_dataset_for_rolling_prediction(
    df, rolling_prediction_length, target_columns, conditional_columns=[], control_columns=[]
):
    # ROLLING_PREDICTION_LENGTH = 192
    context_length = 512
    timestamp_column = "date"
    id_columns = []
    # target_columns = ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"]

    data = df.copy()

    test_end_index = 12 * 30 * 24 + 8 * 30 * 24
    test_start_index = test_end_index - context_length - rolling_prediction_length

    test_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=test_start_index,
        end_index=test_end_index,
    )

    test_dataset = ForecastDFDataset(
        test_data,
        timestamp_column=timestamp_column,
        target_columns=target_columns,
        control_columns=control_columns,
        conditional_columns=conditional_columns,
        prediction_length=rolling_prediction_length,
        context_length=context_length,
    )
    return test_dataset


def test_simple_rolling_prediction(ttm_model, etth_data_base):
    SEED = 42
    rolling_prediction_length = 192
    target_columns = ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"]

    test_dataset = get_dataset_for_rolling_prediction(
        etth_data_base, rolling_prediction_length=rolling_prediction_length, target_columns=target_columns
    )

    model = ttm_model()
    # base_model_context_length = ttm_model.config.context_length
    base_model_prediction_length = model.config.prediction_length

    rec_config = RecursivePredictorConfig(
        model=model,
        requested_prediction_length=rolling_prediction_length,
        model_prediction_length=base_model_prediction_length,
        loss=model.config.loss,
    )
    rolling_model = RecursivePredictor(rec_config)

    temp_dir = tempfile.mkdtemp()
    # zeroshot_trainer
    zeroshot_trainer = Trainer(
        model=rolling_model,
        args=TrainingArguments(
            output_dir=temp_dir, per_device_eval_batch_size=32, seed=SEED, label_names=["future_values"]
        ),
    )

    out = zeroshot_trainer.predict(test_dataset)

    predictions = out.predictions[0]

    assert predictions.shape[2] == len(target_columns)
    assert predictions.shape[1] == rolling_prediction_length


def test_rolling_prediction_with_exogenous(ttm_model, etth_data_base):
    SEED = 42
    rolling_prediction_length = 192
    control_columns = [
        "HUFL",
        "HULL",
        "MUFL",
        "MULL",
        "LUFL",
        "LULL",
    ]
    target_columns = ["OT"]

    # simulate an exogenous model by setting some parameters
    model = ttm_model(prediction_channel_indices=[0], num_input_channels=len(control_columns) + len(target_columns))

    test_dataset = get_dataset_for_rolling_prediction(
        etth_data_base,
        rolling_prediction_length=rolling_prediction_length,
        target_columns=target_columns,
        control_columns=control_columns,
    )

    # base_model_context_length = ttm_model.config.context_length
    base_model_prediction_length = model.config.prediction_length

    rec_config = RecursivePredictorConfig(
        model=model,
        requested_prediction_length=rolling_prediction_length,
        model_prediction_length=base_model_prediction_length,
        loss=model.config.loss,
    )
    rolling_model = RecursivePredictor(rec_config)

    temp_dir = tempfile.mkdtemp()
    # zeroshot_trainer
    zeroshot_trainer = Trainer(
        model=rolling_model,
        args=TrainingArguments(
            output_dir=temp_dir, per_device_eval_batch_size=32, seed=SEED, label_names=["future_values"]
        ),
    )

    out = zeroshot_trainer.predict(test_dataset)

    predictions = out.predictions[0]

    assert predictions.shape[2] == len(target_columns)
    assert predictions.shape[1] == rolling_prediction_length



import logging
from pathlib import Path
from typing import Tuple, Union

import pandas as pd

from .inference_payloads import (
    BaseMetadataInput,
    BaseParameters,
)
from .service_handler import HandlerFunction, ServiceHandler


LOGGER = logging.getLogger(__file__)


class TuningHandler(ServiceHandler):
    @classmethod
    def load(
        cls, model_id: str, model_path: Union[str, Path]
    ) -> Tuple["TuningHandler", None] | Tuple[None, Exception]:
        """Load the handler_config -- the tsfm service config for this model, returning the proper
        handler to use the model.

        model_path is expected to point to a folder containing the tsfm_config.json file. This can be a local folder
        or with a model on the HuggingFace Hub.

        Args:
            model_id (str): A string identifier for the model.
            model_path (Union[str, Path]): The full path to the model, can be a local path or a HuggingFace Hub path.
            handler_function (str): The type of handler, currently supported handlers are defined in the HandlerFunction
                enum.

        """
        # this is loading config and returning
        # the appropriate model handler, it's
        # not loading the model yet
        return super().load(model_id, model_path, handler_function=HandlerFunction.TUNING.value)

    def train(
        self,
        data: pd.DataFrame,
        schema: BaseMetadataInput,
        parameters: BaseParameters,
        tuned_model_name: str,
        tmp_dir: Path,
    ) -> Tuple[str, None] | Tuple[None, Exception]:
        """Perform a fine-tuning request"""
        if not self.prepared:
            return None, RuntimeError("Service wrapper has not yet been prepared; run `handler.prepare()` first.")

        try:
            result = self.implementation.train(
                data, schema=schema, parameters=parameters, tuned_model_name=tuned_model_name, tmp_dir=tmp_dir
            )

            # Does TuneOuput need some info about the request -- for billing purposes?
            return result, None

        except Exception as e:
            return None, e



# Copyright authors of TSPulse: Dual Space Tiny Pre-trained Models for Rapid Time-series Analysis

from typing import Optional, Union

import numpy as np
import torch
from transformers.utils import logging


logger = logging.get_logger(__name__)


def _reduce(metric, reduction="mean", axis=None):
    if reduction == "mean":
        return np.nanmean(metric, axis=axis)
    elif reduction == "sum":
        return np.nansum(metric, axis=axis)
    elif reduction == "none":
        return metric


def mse(
    y: np.ndarray,
    y_hat: np.ndarray,
    reduction: str = "mean",
    axis: Optional[int] = None,
) -> Union[float, np.ndarray]:
    delta_y = np.square(y - y_hat)
    return _reduce(delta_y, reduction=reduction, axis=axis)


def mask_contiguous_with_token(tensor, mask_percentage, patch_length, generator):
    patch_length = 8
    b, s, c = tensor.shape
    num_patches = s // patch_length

    if s % patch_length != 0:
        raise ValueError("Sequence length (s) must be divisible by the patch length (K).")

    tensor = tensor.transpose(-1, -2)  # b, c, s

    num_patches_to_mask_per_sample = int(num_patches * mask_percentage)

    rand_indices = torch.rand((b, c, num_patches), device=tensor.device, generator=generator).argsort(dim=2)

    mask_patches = torch.ones((b, c, num_patches), dtype=torch.bool, device=tensor.device)

    batch_mask = torch.ones(b, dtype=torch.bool, device=tensor.device)

    mask_indices = rand_indices[:, :, :num_patches_to_mask_per_sample]
    mask_patches[batch_mask] = mask_patches[batch_mask].scatter(-1, mask_indices[batch_mask], False)

    patch_mask = mask_patches.unsqueeze(-1).expand(-1, -1, -1, patch_length)

    mask = ~patch_mask.reshape(b, c, s // patch_length, patch_length).permute(0, 2, 3, 1).reshape(b, s, c)

    return mask


def hybrid_masking_with_token(tensor, mask_percentage, patch_size, num_full_patches_to_mask, generator):
    patch_size = 8
    B, T, C = tensor.shape
    device = tensor.device
    total_masks = int(mask_percentage * T)

    if num_full_patches_to_mask * patch_size > total_masks:
        logger.warning(
            f"[hybrid_masking_with_token] num_full_patches_to_mask={num_full_patches_to_mask} "
            f"× patch_size={patch_size} > total_masks={total_masks}. Setting to 0."
        )
        num_full_patches_to_mask = 0

    # === Patch info ===
    patch_ids_full = torch.arange(T, device=device) // patch_size  # [T]
    num_patches = T // patch_size

    rand_vals = torch.rand(B, C, num_patches, device=device, generator=generator)
    _, top_patch_ids = torch.topk(rand_vals, k=num_full_patches_to_mask, dim=2)  # [B, C, K]
    selected_patch_ids = top_patch_ids.unsqueeze(1).expand(B, T, C, num_full_patches_to_mask)
    patch_id_exp = patch_ids_full.view(1, T, 1, 1).expand(B, T, C, num_full_patches_to_mask)
    full_patch_mask = (patch_id_exp == selected_patch_ids).any(-1)  # [B, T, C]

    full_patch_counts = full_patch_mask.sum(dim=1)  # [B, C]
    total_masks_per_channel = int(mask_percentage * T)
    remaining_masks = (total_masks_per_channel - full_patch_counts).clamp(min=0)  # [B, C]

    rand_scores = torch.rand(B, T, C, device=device, generator=generator)
    rand_scores[full_patch_mask] = float("inf")
    sorted_scores, sorted_indices = torch.sort(rand_scores, dim=1)
    time_range = torch.arange(T, device=device).view(1, T, 1).expand(B, T, C)
    topk_mask = time_range < remaining_masks.view(B, 1, C)
    additional_mask = torch.zeros_like(topk_mask)
    additional_mask.scatter_(1, sorted_indices, topk_mask)
    mask = full_patch_mask | additional_mask  # [B, T, C]

    return mask


def mask_generate(generator, x, patch_len, mask_rate, mask_type="hybrid", num_full_patches_for_hybrid_mask=4):
    if mask_type == "hybrid":
        num_full_patches_for_hybrid_mask = int((num_full_patches_for_hybrid_mask) * (mask_rate / 0.125))
        mask = hybrid_masking_with_token(
            x, mask_rate, patch_len, num_full_patches_for_hybrid_mask, generator
        )  # in mask tensor currently missing position are 1
    elif mask_type == "block":
        mask = mask_contiguous_with_token(x, mask_rate, patch_len, generator)
    else:
        raise ValueError("Masking_strategy_not_implemented")

    return mask



import csv
import os
import warnings

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm
from utils import mask_generate, mse

from tsfm_public import TimeSeriesPreprocessor, get_datasets
from tsfm_public.models.tspulse import TSPulseForReconstruction


warnings.filterwarnings("ignore")

device = "cuda:0" if torch.cuda.is_available() else "cpu"
CONTEXT_LEN = 512
FORECAST_LEN = 0


def main(DATASET, mask_type, mask_ratio):
    if DATASET in ["ETTh1", "ETTh2", "ETTm1", "ETTm2"]:
        batch_size = 64
    else:
        batch_size = 4

    # Dataset
    if DATASET in ["ETTh1", "ETTh2", "ETTm1", "ETTm2"]:
        dataset_path = f"https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/{DATASET}.csv"
    else:
        dataset_path = f"datasets/{DATASET}/{DATASET}.csv"

    timestamp_column = "date"
    id_columns = []

    if DATASET in ["ETTh1", "ETTh2"]:
        split_config = {
            "train": [0, 8640],
            "valid": [8640, 11520],
            "test": [
                11520,
                14400,
            ],
        }
    elif DATASET in ["ETTm1", "ETTm2"]:
        split_config = {
            "train": [0, 34560],
            "valid": [34560, 46080],
            "test": [
                46080,
                57600,
            ],
        }
    else:
        split_config = {
            "train": 0.7,
            "test": 0.2,
        }

    data = pd.read_csv(
        dataset_path,
        parse_dates=[timestamp_column],
    )

    target_columns = data.columns.to_list()[1:]  # all the columns from the data except 'date'

    column_specifiers = {
        "timestamp_column": timestamp_column,
        "id_columns": id_columns,
        "target_columns": target_columns,
        "control_columns": [],
    }

    tsp = TimeSeriesPreprocessor(
        **column_specifiers,
        context_length=CONTEXT_LEN,
        prediction_length=FORECAST_LEN,
        scaling=True,
        encode_categorical=False,
        scaler_type="standard",
    )

    _, _, dset_test = get_datasets(tsp, data, split_config)

    def collate_only_past_values(batch):
        return torch.stack([item["past_values"] for item in batch])

    test_dataloader = DataLoader(dset_test, batch_size=batch_size, shuffle=False, collate_fn=collate_only_past_values)

    model = TSPulseForReconstruction.from_pretrained(
        "ibm-granite/granite-timeseries-tspulse-r1",
        revision="tspulse-hybrid-dualhead-512-p8-r1",
        num_input_channels=tsp.num_input_channels,
        mask_type="user",
    ).to(device)

    seed = 42
    g = torch.Generator(device=device)
    g.manual_seed(seed)

    trues, preds, masks = [], [], []
    with torch.no_grad():
        for batch in tqdm(test_dataloader):
            batch_x = batch.to(device)  # b l c

            mask = mask_generate(g, batch_x, 8, mask_ratio, mask_type)

            output = model(past_values=batch_x, past_observed_mask=~mask)

            reconstructed_output = output.reconstruction_outputs

            trues.append(batch_x.detach().cpu().numpy())
            preds.append(reconstructed_output.detach().cpu().numpy())
            masks.append(mask.detach().cpu().numpy())

        preds = np.concatenate(preds)
        trues = np.concatenate(trues)
        masks = np.concatenate(masks)

        MSE = mse(y=trues[masks == 1], y_hat=preds[masks == 1], reduction="mean")
        print(f"Dataset = {DATASET}  : Mask Type = {mask_type}  : Mask Ratio = {mask_ratio}")
        print(f"Mean Squarred Error (MSE)={MSE:.3f}")

        output_file = "tspulse_zeroshot_imputation_results.csv"

        if not os.path.exists(output_file):
            with open(output_file, mode="w", newline="") as file:
                writer = csv.writer(file)
                writer.writerow(["Dataset", "Mask Type", "Mask Ratio", "MSE"])

        with open(output_file, mode="a", newline="") as file:
            writer = csv.writer(file)
            writer.writerow([DATASET, mask_type, mask_ratio, f"{MSE:.3f}"])


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--datasets",
        nargs="+",
        type=str,
        required=True,
        choices=["ETTh1", "ETTh2", "ETTm1", "ETTm2", "weather", "electricity"],
        help="List of UEA dataset names",
    )
    parser.add_argument(
        "--mask_type",
        nargs="+",
        type=str,
        required=True,
        choices=["block", "hybrid"],
        help="Masking strategy to evaluate. Options available : 'block' and 'hybrid'",
    )
    parser.add_argument(
        "--mask_ratios",
        nargs="+",
        type=float,
        required=True,
        choices=[0.125, 0.25, 0.375, 0.5],
        help="Masking ratios to evaluate. Options available : 0.125, 0.25, 0.375, 0.5",
    )
    args = parser.parse_args()

    for m_t in args.mask_type:
        for DATASET in args.datasets:
            for m_r in args.mask_ratios:
                main(DATASET, m_t, m_r)



# Copyright contributors to the TSFM project
#
"""Helper class for anomaly detection support"""

from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import List, Union

import numpy as np
import pandas as pd
import torch
from transformers.utils.generic import ModelOutput


ScoreType = Union[np.ndarray, torch.Tensor]
ScoreListType = Union[ScoreType, List[ScoreType]]


class AnomalyScoreMethods(Enum):
    """Enum type for time series foundation model based anomaly detection modes."""

    MEAN_DEVIATION = "meandev"
    PREDICTIVE = "forecast"
    TIME_RECONSTRUCTION = "time"
    FREQUENCY_RECONSTRUCTION = "fft"
    PROBABILISTIC = "probabilistic"


class TSADHelperUtility:
    """Abstract class for Anomaly detection pipeline.
    Model specific implementation. Implements three API calls required for integration to Anomaly Detection Pipeline.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def is_valid_mode(self, mode_str: List[str]) -> bool:
        """Validates given Anomaly Prediction mode is supported by the model or not.

        Args:
            mode_str (list): list of prediction mode string

        Returns:
            bool: returns true if supported
        """
        raise NotImplementedError

    def preprocess(self, x: pd.DataFrame, **kwargs) -> pd.DataFrame:
        """API to support model specific data preprocessing

        Args:
            x (pd.DataFrame): input to the pipeline

        Returns:
            pd.DataFrame: processed data frame
        """
        return x

    @abstractmethod
    def compute_score(
        self,
        payload: dict,
        **kwargs,
    ) -> ModelOutput:
        """Invokes model to generate output required for anomaly score computation.

        Args:
            payload (dict): data batch

        Returns:
            ModelOutput: return model outputs
        """
        raise NotImplementedError

    @abstractmethod
    def adjust_boundary(
        self,
        key: str,
        x: ScoreListType,
        **kwargs,
    ) -> np.ndarray:
        """API to adjust scores at the data boundary.

        Args:
            key (str): key associated with the model output
            x (ScoreListType): full model outputs

        Returns:
            np.ndarray: combined scores
        """
        raise NotImplementedError



"""Inference handler for TSFM models"""

import copy
import logging
import tempfile
from functools import cache
from pathlib import Path
from typing import Any, Dict, Optional, Union

import pandas as pd
import torch

from tsfm_public import TimeSeriesForecastingPipeline
from tsfm_public.toolkit.time_series_preprocessor import TimeSeriesPreprocessor, extend_time_series
from tsfm_public.toolkit.tsfm_config import TSFMConfig
from tsfm_public.toolkit.util import select_by_index

from . import TSFM_ALLOW_LOAD_FROM_HF_HUB
from .inference_payloads import ForecastingMetadataInput, ForecastingParameters
from .tsfm_util import load_config, load_model, register_config


LOCAL_FILES_ONLY = not TSFM_ALLOW_LOAD_FROM_HF_HUB
LOGGER = logging.getLogger(__file__)


class TSFMForecastingInferenceHandler:
    def __init__(
        self,
        model_id: str,
        model_path: Union[str, Path],
        handler_config: TSFMConfig,
    ):
        if (
            getattr(handler_config, "model_type", None)
            and getattr(handler_config, "model_config_name", None)
            and getattr(handler_config, "module_path", None)
        ):
            register_config(
                handler_config.model_type,
                handler_config.model_config_name,
                handler_config.module_path,
            )
            LOGGER.info(f"registered {handler_config.model_type}")

        self.model_id = model_id
        self.model_path = model_path
        self.handler_config = handler_config

        # set during prepare
        self.config = None
        self.model = None
        self.preprocessor = None

        # loosen the schema checking when using a saved preprocessor
        self.strict_schema_match = False

    def _get_config_kwargs(
        self,
        parameters: Optional[ForecastingParameters] = None,
        preprocessor: Optional[TimeSeriesPreprocessor] = None,
    ) -> Dict[str, Any]:
        """Helper function to return additional configuration arguments that are used during config load.
        Can be overridden in a subclass to allow specialized model functionality.

        Args:
            parameters (Optional[ForecastingParameters], optional): Request parameters. Defaults to None.
            preprocessor (Optional[TimeSeriesPreprocessor], optional): Time seres preprocessor. Defaults to None.

        Returns:
            Dict[str, Any]: Dictionary of additional arguments that are used later as keyword arguments to the config.
        """
        return {"num_input_channels": preprocessor.num_input_channels}

    @classmethod
    @cache
    def _cached_load_model(cls, model_path, config: str, module_path, config_class):
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".json", delete=True) as tmp:
            tmp.write(config)
            tmp.flush()
            return load_model(
                model_path=model_path,
                config=config_class.from_json_file(tmp.name),
                module_path=module_path,
            )

    def prepare(
        self,
        data: pd.DataFrame,
        future_data: Optional[pd.DataFrame] = None,
        schema: Optional[ForecastingMetadataInput] = None,
        parameters: Optional[ForecastingParameters] = None,
        **kwargs,
    ) -> "TSFMForecastingInferenceHandler":
        """Implementation of prepare for HF-like models. We assume the model will make use of the TSFM
        preprocessor and forecasting pipeline. This method:
        1) loades the preprocessor, creating a new one if the model does not already have a preprocessor
        2) updates model configuration arguments by calling _get_config_kwargs
        3) loads the HuggingFace model, passing the updated config object

        Args:
            data (pd.DataFrame): A pandas dataframe containing historical data.
            future_data (Optional[pd.DataFrame], optional): A pandas dataframe containing future data. Defaults to None.
            schema (Optional[ForecastingMetadataInput], optional): Schema information from the original inference
                request. Includes information about columns and their role. Defaults to None.
            parameters (Optional[ForecastingParameters], optional): Parameters from the original inference
                request. Defaults to None.

        Returns:
            ForecastingHuggingFaceHandler: The updated service handler object.
        """

        preprocessor_params = copy.deepcopy(schema.model_dump())
        preprocessor_params["prediction_length"] = parameters.prediction_length

        LOGGER.info(f"Preprocessor params: {preprocessor_params}")

        # load preprocessor
        try:
            preprocessor = TimeSeriesPreprocessor.from_pretrained(self.model_path, local_files_only=LOCAL_FILES_ONLY)
            LOGGER.info("Successfully loaded preprocessor")
        except OSError:
            preprocessor = None
            LOGGER.info("No preprocessor found")
        except Exception as ex:
            raise ex

        if self.handler_config.is_finetuned and preprocessor is None:
            raise ValueError("Model indicates that it is finetuned but no preprocessor was found.")

        if not self.handler_config.is_finetuned and preprocessor is not None:
            raise ValueError("Unexpected: model indicates that it is not finetuned but a preprocessor was found.")

        if preprocessor is None:
            to_check = [
                "conditional_columns",
                "control_columns",
                "observable_columns",
                "static_categorical_columns",
                "categorical_columns",
            ]

            for param in to_check:
                if param in preprocessor_params and preprocessor_params[param]:
                    raise ValueError(
                        f"Unexpected parameter {param} for a zero-shot model, please confirm you have the correct model_id and schema."
                    )

            preprocessor = TimeSeriesPreprocessor(
                **preprocessor_params,
                scaling=False,
                encode_categorical=False,
            )
            # train to estimate freq
            preprocessor.train(data)
            LOGGER.info(f"Data frequency determined: {preprocessor.freq}")
        else:
            # check payload, but only certain parameters
            # id_columns are not checked, since an artificial id may be added to do batch inference
            to_check_columns = [
                "timestamp_column",
                "target_columns",
                "conditional_columns",
                "control_columns",
                "observable_columns",
                "categorical_columns",
                "static_categorical_columns",
            ]

            to_check_params = ["freq"]
            to_check = to_check_columns.copy()
            to_check.extend(to_check_params)

            for param in to_check:
                param_val_saved = getattr(preprocessor, param)
                param_val = preprocessor_params[param]
                # if a parameter is passed we check it matches the preprocessor
                if self.strict_schema_match or not (param_val is None or param_val == [] or param_val == ""):
                    if param_val != param_val_saved:
                        raise ValueError(
                            f"Attempted to use a fine-tuned model with a different schema, please confirm you have the correct model_id and schema. Error in parameter {param}: received {param_val} but expected {param_val_saved}."
                        )
                # then we check that parameters from the saved preprocessor match what is in the data
                if param in to_check_params:
                    continue
                p = param_val_saved if isinstance(param_val_saved, list) else [param_val_saved]
                if any(c not in data.columns for c in p):
                    raise ValueError(
                        f"Attempted to use a fine-tuned model with data that does not match the saved schema, please confirm you have the correct model_id and appropriate data. Error in parameter {param}: data does not contain a column named {param_val_saved}."
                    )

        model_config_kwargs = self._get_config_kwargs(
            parameters=parameters,
            preprocessor=preprocessor,
        )
        LOGGER.info(f"model_config_kwargs: {model_config_kwargs}")
        model_config = load_config(self.model_path, **model_config_kwargs)

        model = TSFMForecastingInferenceHandler._cached_load_model(
            self.model_path,
            config=model_config.to_json_string(),
            module_path=self.handler_config.module_path,
            config_class=model_config.__class__,
        )

        self.config = model_config
        self.model = model
        self.preprocessor = preprocessor

    def run(
        self,
        data: pd.DataFrame,
        future_data: Optional[pd.DataFrame] = None,
        schema: Optional[ForecastingMetadataInput] = None,
        parameters: Optional[ForecastingParameters] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Implementation of run for TSFM models.
        Checks prediction length, data length (both past and future exogenous), configures batch size, determines
        device and uses the forecasting pipeline to generate forecasts.

        Args:
            data (pd.DataFrame): Input historical time series data.
            future_data (Optional[pd.DataFrame], optional): Input future time series data, useful for
                passing future exogenous if known. Defaults to None.
            schema (Optional[ForecastingMetadataInput], optional): Schema information from the original inference
                request. Includes information about columns and their role. Defaults to None.
            parameters (Optional[ForecastingParameters], optional): Parameters from the original inference
                request. Defaults to None.

        Returns:
            pd.DataFrame: The forecasts produced by the model.
        """

        # warn if future data is not provided, but is needed by the model
        # Remember preprocessor.exogenous_channel_indices are the exogenous for which future data is available
        if self.preprocessor.exogenous_channel_indices and future_data is None:
            raise ValueError(
                "Future data should be provided for exogenous columns where the future is known (`control_columns` and `observable_columns`)"
            )

        if not self.preprocessor.exogenous_channel_indices and future_data is not None:
            raise ValueError("Future data was provided, but the model does not support or require future exogenous.")

        # future_data checks
        if future_data is not None:
            if schema.id_columns:
                data_lengths = future_data.groupby(schema.id_columns)[schema.id_columns].apply(len)
                fd_min_len_index = data_lengths.argmin()
                fd_min_data_length = data_lengths.iloc[fd_min_len_index]
                fd_max_data_length = data_lengths.max()
            else:
                fd_min_data_length = fd_max_data_length = len(future_data)
            LOGGER.info(
                f"Future Data length recieved {len(future_data)}, minimum series length: {fd_min_data_length}, maximum series length: {fd_max_data_length}"
            )

            # if data is too short, raise error
            prediction_length = getattr(self.config, "prediction_filter_length", None)
            has_prediction_filter = prediction_length is not None

            model_prediction_length = self.config.prediction_length

            prediction_length = prediction_length if prediction_length is not None else model_prediction_length
            if fd_min_data_length < prediction_length:
                err_str = (
                    "Future data should have time series of length that is at least the specified prediction length. "
                )
                if schema.id_columns:
                    err_str += f"Received {fd_min_data_length} time points for id {data_lengths.index[fd_min_len_index]}, but expected {prediction_length} time points."
                else:
                    err_str += (
                        f"Received {fd_min_data_length} time points, but expected {prediction_length} time points."
                    )
                raise ValueError(err_str)

            # if data exceeds prediction filter length, truncate
            if fd_max_data_length > model_prediction_length:
                LOGGER.info(f"Truncating future series lengths to {model_prediction_length}")
                future_data = select_by_index(
                    future_data, id_columns=schema.id_columns, end_index=model_prediction_length
                )

            # if provided data is greater than prediction_filter_length, but less than model_prediction_length we extend
            if has_prediction_filter and fd_min_data_length < model_prediction_length:
                LOGGER.info(f"Extending time series to match model prediction length: {model_prediction_length}")
                future_data = extend_time_series(
                    time_series=future_data,
                    freq=self.preprocessor.freq,
                    timestamp_column=schema.timestamp_column,
                    grouping_columns=schema.id_columns,
                    total_periods=model_prediction_length,
                )

        batch_size = (
            parameters.inference_batch_size
            if parameters.inference_batch_size
            else self.handler_config.inference_batch_size
        )
        LOGGER.info(f"Using inference batch size: {batch_size}")

        device = "cpu" if not torch.cuda.is_available() else "cuda"

        extra_pipeline_args = getattr(self.handler_config, "extra_pipeline_arguments", {})
        forecast_pipeline = TimeSeriesForecastingPipeline(
            model=self.model,
            explode_forecasts=True,
            feature_extractor=self.preprocessor,
            add_known_ground_truth=False,
            freq=self.preprocessor.freq,
            device=device,
            batch_size=1000,
            **extra_pipeline_args,
        )
        forecasts = forecast_pipeline(data, future_time_series=future_data, inverse_scale_outputs=True)

        return forecasts

    def calculate_data_point_counts(
        self,
        data: pd.DataFrame,
        future_data: Optional[pd.DataFrame] = None,
        output_data: Optional[pd.DataFrame] = None,
        schema: Optional[ForecastingMetadataInput] = None,
        parameters: Optional[ForecastingParameters] = None,
    ) -> Dict[str, int]:
        """Implementation for counting datapoints in input and output

        Assumes data has been truncated
        Future data may not be truncated
        """

        input_ts_columns = sum(
            [
                len(c)
                for c in [
                    schema.target_columns,
                    schema.conditional_columns,
                    schema.control_columns,
                    schema.observable_columns,
                ]
            ]
        )
        input_ts_columns = input_ts_columns if input_ts_columns != 0 else data.shape[1] - len(schema.id_columns) - 1
        input_static_columns = len(schema.static_categorical_columns)
        num_target_columns = (
            len(schema.target_columns) if schema.target_columns != [] else data.shape[1] - len(schema.id_columns) - 1
        )
        unique_ts = len(data.drop_duplicates(subset=schema.id_columns)) if schema.id_columns else 1
        has_future_data = future_data is not None

        # we don't count the static columns in the future data
        # we only count future data which falls within forecast horizon "causal assumption"
        # note that output_data.shape[0] = unique_ts * prediction_length
        future_data_points = (input_ts_columns - num_target_columns) * output_data.shape[0] if has_future_data else 0

        counts = {
            "input_data_points": input_ts_columns * data.shape[0]
            + input_static_columns * unique_ts
            + future_data_points,
            "output_data_points": output_data.shape[0] * num_target_columns,
        }
        LOGGER.info(f"Data point counts: {counts}")
        return counts


class TinyTimeMixerForecastingInferenceHandler(TSFMForecastingInferenceHandler):
    """Service handler for the tiny time mixer model"""

    def _get_config_kwargs(
        self,
        parameters: Optional[ForecastingParameters] = None,
        preprocessor: Optional[TimeSeriesPreprocessor] = None,
    ) -> Dict[str, Any]:
        """Implements the _get_config_kwargs method for TTM models.

        Args:
            parameters (Optional[ForecastingParameters], optional): Inference parameters. Defaults to None.
            preprocessor (Optional[TimeSeriesPreprocessor], optional): Configured preprocessor. Defaults to None.

        Returns:
            Dict[str, Any]: Updated config arguments that get passed to during model load.
        """
        config_kwargs = {
            "num_input_channels": preprocessor.num_input_channels,
            "prediction_filter_length": parameters.prediction_length,
            "exogenous_channel_indices": preprocessor.exogenous_channel_indices,
            "prediction_channel_indices": preprocessor.prediction_channel_indices,
        }
        return config_kwargs



# Copyright contributors to the TSFM project
#

"""Tests the time series classification preprocessor and functions"""

import numpy as np
import pandas as pd
import pytest

from tsfm_public.toolkit.time_series_classification_preprocessor import (
    TimeSeriesClassificationPreprocessor,
    check_nested_lengths,
    nest_transform,
    unnest_transform,
)


def test_label_encodes(ts_data_nested):
    df = ts_data_nested.copy()

    tsp = TimeSeriesClassificationPreprocessor(
        timestamp_column="time_date",
        input_columns=["val", "val2"],
        label_column="label",
        id_columns=[
            "id",
        ],
    )

    tsp.train(df)
    assert tsp.label_encoder is not None
    assert len(tsp.label_encoder.classes_) == len(df.label.unique())

    df_prep = tsp.preprocess(df)
    assert df_prep.label.dtype == "int64"


def test_scaling(ts_data_nested):
    df = ts_data_nested.copy()

    tsp = TimeSeriesClassificationPreprocessor(
        timestamp_column="time_date",
        input_columns=["val", "val2"],
        label_column="label",
        id_columns=[
            "id",
        ],
        scaling=True,
    )

    tsp.train(df)

    df_prep = tsp.preprocess(df)

    # to do add condition
    np.testing.assert_almost_equal(df_prep.val.apply(np.mean).mean(), 0)


def test_static_categorical(ts_data_nested):
    df = ts_data_nested.copy()
    df["cat"] = df.label.copy()

    tsp = TimeSeriesClassificationPreprocessor(
        timestamp_column="time_date",
        input_columns=["val", "val2"],
        label_column="label",
        static_categorical_columns=["cat"],
        id_columns=[
            "id",
        ],
        scaling=True,
    )

    with pytest.raises(Exception):
        tsp.categorical_vocab_size_list

    tsp.train(df)
    assert tsp.categorical_vocab_size_list == [4]

    df_prep = tsp.preprocess(df)
    assert df_prep["cat"].dtype == "float64"


def test_helpers():
    tsp = TimeSeriesClassificationPreprocessor(
        timestamp_column="time_date",
        input_columns=["val", "val2"],
        label_column="label",
        id_columns=[
            "id",
        ],
        scaling_id_columns=["id"],
    )

    assert tsp.num_input_channels == 2
    assert tsp.exogenous_channel_indices == []
    assert tsp.categorical_vocab_size_list is None


def test_unnest_nest(ts_data_nested):
    df = ts_data_nested.copy()

    u = unnest_transform(df, columns=["val", "val2"])
    n = nest_transform(u, columns=["val", "val2"])

    # check length
    assert df.shape[0] == n.shape[0]

    # check all nested rows
    for i in range(df.shape[0]):
        assert np.all(df["val"].iloc[i] == n["val"].iloc[i])
        assert np.all(df["val2"].iloc[i] == n["val2"].iloc[i])


def test_check_nested_length(ts_data_nested):
    df = ts_data_nested.copy()
    check_nested_lengths(df, columns=["val", "val2"])

    # df.val.loc[0] = pd.Series(np.concatenate((df["val"].iloc[0].values, [10])))

    row = df.iloc[0].copy()
    row.val = pd.Series(np.concatenate((df["val"].iloc[0].values, [10])))
    df_new = pd.concat([df, pd.DataFrame(row).T], ignore_index=True)

    with pytest.raises(Exception):
        check_nested_lengths(df_new, columns=["val", "val2"])

    check_nested_lengths(df, columns=["val"])

    row.val2 = pd.Series(np.concatenate((df["val2"].iloc[0].values, [-10])))
    df_new = pd.concat([df, pd.DataFrame(row).T], ignore_index=True)
    check_nested_lengths(df_new, columns=["val", "val2"])



"""Tests for util functions"""

import tempfile

import pandas as pd
import pytest

from tsfm_public.toolkit.util import convert_to_univariate, convert_tsfile, get_split_params, train_test_split


split_cases = [
    (0, 1, "select_by_index"),
    (0, 0.1, "select_by_relative_fraction"),
    (0.0, 0.1, "select_by_relative_fraction"),
    (0.0, 200.0, "select_by_index"),
    (0.0, 200, "select_by_index"),
    (0.5, 1, "select_by_relative_fraction"),
    (0.5, 1.0, "select_by_relative_fraction"),
    (10, 100.0, "select_by_index"),
]


@pytest.mark.parametrize("left_arg,right_arg,expected", split_cases)
def test_get_split_params(left_arg, right_arg, expected):
    """Test that get_split_params gives the right split function"""

    split_config, split_function = get_split_params({"train": [left_arg, right_arg], "valid": [0, 1], "test": [0, 1]})

    assert split_function["train"].__name__ == expected


def test_train_test_split():
    n = 100
    df = pd.DataFrame({"date": range(n), "value": range(n)})

    train, valid, test = train_test_split(df, train=0.7, test=0.2)

    assert len(train) == int(n * 0.7)
    assert len(test) == int(n * 0.2)
    valid_len_100 = n - int(n * 0.7) - int(n * 0.2)
    assert len(valid) == valid_len_100

    n = 101
    df = pd.DataFrame({"date": range(n), "value": range(n)})

    train, valid, test = train_test_split(df, train=0.7, test=0.2)

    assert len(train) == int(n * 0.7)
    assert len(test) == int(n * 0.2)
    valid_len_101 = n - int(n * 0.7) - int(n * 0.2)
    assert len(valid) == valid_len_101

    assert valid_len_100 + 1 == valid_len_101


def test_convert_to_univariate(ts_data):
    id_columns = ["id"]
    timestamp_column = "timestamp"
    target_columns = ["value1", "value2"]
    df_uni = convert_to_univariate(
        ts_data, timestamp_column=timestamp_column, id_columns=id_columns, target_columns=target_columns
    )

    assert df_uni.columns.to_list() == [
        timestamp_column,
    ] + id_columns + ["column_id", "value"]

    assert len(df_uni) == len(target_columns) * len(ts_data)

    # test single column

    target_columns = ["value1"]

    with pytest.raises(ValueError):
        df_uni = convert_to_univariate(
            ts_data, timestamp_column=timestamp_column, id_columns=id_columns, target_columns=target_columns
        )


def test_convert_tsfile():
    data = """#Test
#
#The classes are
#1. one
#2. two
#3. three
@problemName test
@timeStamps false
@missing false
@univariate true
@equalLength true
@seriesLength 5
@classLabel true 1 2 3
@data
1,2,3,4,5:1
10,20,30,40,50:2
11,12,13,14,15:3
"""
    with tempfile.NamedTemporaryFile() as t:
        t.write(data.encode("utf-8"))
        t.flush()
        df = convert_tsfile(t.name)

    assert df.shape == (15, 3)



import argparse

import faiss
import numpy as np
import pandas as pd
import torch
from datautils import (
    RetrievedData,
    SyntheticCombDataset,
    TransformedSyntheticTestDataset,
    UCRDataset,
)
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import set_seed

from tsfm_public.models.tspulse import TSPulseForReconstruction
from tsfm_public.models.tspulse.utils.helpers import get_embeddings


@torch.no_grad()
def retrieve(device, batch_size, k, train_dataset, test_dataset, model):
    # compute embeddings of training data
    dataloader = DataLoader(train_dataset, batch_size=batch_size)
    train_embeddings, metadata = [], []
    for batch_idx, batch in enumerate(tqdm(dataloader)):
        embs = get_embeddings(model, batch["past_values"].to(device))
        B, C, D = embs.shape
        metadata.extend([(batch_idx * batch_size + b, c) for b in range(B) for c in range(C)])
        train_embeddings.append(embs.cpu().numpy())
    train_embeddings = np.concatenate(train_embeddings).squeeze(axis=1)  # due to univariate

    # create index set of embeddings from training data
    d = train_embeddings.shape[1]
    quantizer = faiss.IndexFlatL2(d)
    index = faiss.IndexIVFFlat(quantizer, d, 20)
    index.train(train_embeddings)
    index.add(train_embeddings)
    index.nprobe = 5

    # find top-k similar embeedings for query test data
    dataloader = DataLoader(test_dataset, batch_size=batch_size)
    D_all, I_all = [], []
    for batch_idx, batch in enumerate(tqdm(dataloader)):
        test_embedding = get_embeddings(model, batch["past_values"].to(device))
        query_vector = test_embedding.squeeze(dim=1).cpu()  # due to univariate, [B, 1, D] -> [B, D]
        D, I = index.search(query_vector, k=k)
        D_all.extend(D)
        I_all.extend(I)

    return D_all, I_all, metadata


def compute_ranking_score(device, batch_size, k, I_all, train_dataset, test_dataset, level):
    def _precision_k(cmp, k):
        cmp_k = np.sum(cmp[:, :k], axis=1) / k
        mean_cmp_k = np.mean(cmp_k)
        return mean_cmp_k

    def _avg_prec_k(cmp, n_rel, k):
        rank = np.arange(k) + 1
        cmp_k = np.cumsum(cmp[:, :k], axis=1)
        cmp_k = cmp_k / rank
        cmp_k = np.sum(cmp_k * cmp[:, :k], axis=1) / np.minimum(n_rel, k)
        mean_cmp_k = np.mean(cmp_k)
        return mean_cmp_k

    def _dcg_k(cmp, k):
        cmp_k = cmp[:, :k]
        n_data = cmp_k.shape[0]
        dcgs = np.zeros(n_data)
        for i in range(n_data):
            idx = np.argwhere(cmp_k[i, :] > 0).flatten()
            dcg = np.reciprocal(np.log2(idx + 2))
            dcgs[i] = np.sum(dcg) / k
        return dcgs

    def _idcg_k(cmp, n_rel, k):
        n_data = cmp.shape[0]
        ideal_cmp = np.zeros((n_data, k))
        n_rel_each = np.minimum(k, n_rel)
        mask = np.repeat((np.arange(k) + 1)[:, None].T, n_data, axis=0)
        mask = mask <= n_rel_each[:, None]
        ideal_cmp[mask] = 1
        return _dcg_k(ideal_cmp, k)

    def _mrr(cmp, k):
        cmp_k = cmp[:, :k]
        first_index = np.argmax(cmp_k, axis=1, keepdims=True)
        first_value = np.take_along_axis(cmp_k, first_index, axis=1)  # false if there is no correct item
        rr = 1 / (first_index + 1)
        return np.mean(rr * first_value)

    retrieveddata = RetrievedData(train_dataset, test_dataset, I_all, level)
    dataloader = DataLoader(retrieveddata, batch_size=batch_size)
    cmp, n_rel = [], []
    for batch_idx, batch in enumerate(tqdm(dataloader)):
        label_test = batch["label_test"]
        labels_train = batch["labels_train"]
        cmp.append(label_test[:, None] == labels_train)
        n_rel.append(batch["n_rel"])
    cmp = np.concatenate(cmp, axis=0)
    n_rel = np.concatenate(n_rel, axis=0)

    # compute metrics
    prec_k = _precision_k(cmp, k)
    avg_prec_k = _avg_prec_k(cmp, n_rel, k)
    dcg_k = _dcg_k(cmp, k)
    idcg_k = _idcg_k(cmp, n_rel, k)
    ndcg_k = np.mean(dcg_k / idcg_k)
    mrr_k = _mrr(cmp, k)

    return prec_k, avg_prec_k, ndcg_k, mrr_k


def evaluate(device, seed, batch_size, k, model, train_dataset, test_dataset):
    D_all, I_all, metadata = retrieve(device, batch_size, k, train_dataset, test_dataset, model)
    df = []
    for level in ["family_match", "finegrained_match"]:
        prec_k, avg_prec_k, ndcg_k, mrr_k = compute_ranking_score(
            device, batch_size, k, I_all, train_dataset, test_dataset, level
        )
        df.append(
            pd.DataFrame(
                {
                    "PREC@k": prec_k,
                    "AP@k": avg_prec_k,
                    "NDCG@k": ndcg_k,
                    "MRR@k": mrr_k,
                    "k": k,
                    "seed": seed,
                    "level": level,
                },
                index=[0],
            )
        )

    df = pd.concat(df, axis=0).reset_index(drop=True)

    return df


def main(seed, batch_size, k, dataset):
    set_seed(seed)
    device = torch.cuda.current_device() if torch.cuda.is_available() else "cpu"

    if dataset == "synth":
        train_dataset = SyntheticCombDataset()
    elif dataset == "real":
        train_dataset = UCRDataset()
    else:
        raise RuntimeError(f"{dataset} is not found")

    batch_size = min([batch_size, len(train_dataset)])
    test_dataset = TransformedSyntheticTestDataset(train_dataset, max_shift=0.2, max_scale=0.2, noise_ratio=0.1)

    model = TSPulseForReconstruction.from_pretrained(
        "ibm-granite/granite-timeseries-tspulse-r1",
        revision="tspulse-hybrid-dualhead-512-p8-r1",
        num_input_channels=1,
        mask_type="user",
    )
    model.eval().to(device)

    df_result = evaluate(device, seed, batch_size, k, model, train_dataset, test_dataset)
    return df_result


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("dataset", choices=["synth", "real"])
    args = parser.parse_args()

    dataset = args.dataset
    batch_size = 1024
    k = 3
    seed = 42
    df_result = main(seed, batch_size, k, dataset)
    df_result.to_csv(f"results_{dataset}.csv")
    print(df_result[["level", "PREC@k", "MRR@k", "AP@k", "NDCG@k"]])



# Copyright contributors to the TSFM project
#
"""Utilities for TTM notebooks"""

import argparse
import logging
import os
import tempfile

import torch


logger = logging.getLogger(__name__)


def get_ttm_args():  # pragma: no cover
    parser = argparse.ArgumentParser(description="TTM pretrain arguments.")
    # Adding a positional argument
    parser.add_argument(
        "--forecast_length",
        "-fl",
        type=int,
        required=False,
        default=96,
        help="Forecast length",
    )
    parser.add_argument(
        "--context_length",
        "-cl",
        type=int,
        required=False,
        default=512,
        help="History context length",
    )
    parser.add_argument(
        "--patch_length",
        "-pl",
        type=int,
        required=False,
        default=64,
        help="Patch length",
    )
    parser.add_argument(
        "--adaptive_patching_levels",
        "-apl",
        type=int,
        required=False,
        default=3,
        help="Number of adaptive patching levels of TTM",
    )
    parser.add_argument(
        "--d_model_scale",
        "-dms",
        type=int,
        required=False,
        default=3,
        help="Model hidden dimension",
    )
    parser.add_argument(
        "--decoder_d_model_scale",
        "-ddms",
        type=int,
        required=False,
        default=2,
        help="Decoder hidden dimension",
    )
    parser.add_argument(
        "--num_gpus",
        "-ng",
        type=int,
        required=False,
        default=None,
        help="Number of GPUs",
    )
    parser.add_argument("--random_seed", "-rs", type=int, required=False, default=42, help="Random seed")
    parser.add_argument("--batch_size", "-bs", type=int, required=False, default=3000, help="Batch size")
    parser.add_argument(
        "--num_epochs",
        "-ne",
        type=int,
        required=False,
        default=25,
        help="Number of epochs",
    )

    parser.add_argument(
        "--num_workers",
        "-nw",
        type=int,
        required=False,
        default=8,
        help="Number of dataloader workers",
    )
    parser.add_argument(
        "--learning_rate",
        "-lr",
        type=float,
        required=False,
        default=0.001,
        help="Learning rate",
    )
    parser.add_argument(
        "--dataset",
        "-ds",
        type=str,
        required=False,
        default="etth1",
        help="Dataset",
    )
    parser.add_argument(
        "--data_root_path",
        "-drp",
        type=str,
        required=False,
        default="datasets/",
        help="Dataset",
    )
    parser.add_argument(
        "--save_dir",
        "-sd",
        type=str,
        required=False,
        default=tempfile.gettempdir(),
        help="Data path",
    )
    parser.add_argument(
        "--early_stopping",
        "-es",
        type=int,
        required=False,
        default=1,
        help="Whether to use early stopping during finetuning.",
    )
    parser.add_argument(
        "--freeze_backbone",
        "-fb",
        type=int,
        required=False,
        default=1,
        help="Whether to freeze the backbone during few-shot finetuning.",
    )
    parser.add_argument(
        "--enable_prefix_tuning",
        "-ept",
        type=int,
        required=False,
        default=0,
        help="Enable prefix tuning in TTM.",
    )
    parser.add_argument(
        "--hf_model_path",
        "-hmp",
        type=str,
        required=False,
        default="ibm-granite/granite-timeseries-ttm-r2",
        help="Hugginface model card path.",
    )

    parser.add_argument(
        "--datasets",
        type=str,
        default=None,
        help="List of datasets, comma separated",
    )

    parser.add_argument(
        "--zeroshot",
        "-zs",
        type=int,
        required=False,
        default=1,
        help="Run Zeroshot",
    )

    parser.add_argument(
        "--fewshot",
        "-fs",
        type=int,
        required=False,
        default=1,
        help="Run Zeroshot",
    )

    parser.add_argument(
        "--plot",
        type=int,
        required=False,
        default=1,
        help="Enable plotting",
    )

    parser.add_argument(
        "--num_layers",
        type=int,
        required=False,
        default=2,
        help="Number of  layers",
    )

    parser.add_argument(
        "--decoder_num_layers",
        type=int,
        required=False,
        default=2,
        help="Number of  decoder layers",
    )

    parser.add_argument(
        "--dropout",
        type=float,
        required=False,
        default=0.2,
        help="Dropout",
    )

    parser.add_argument(
        "--head_dropout",
        type=float,
        required=False,
        default=0.2,
        help="head_dropout",
    )

    # Parsing the arguments
    args = parser.parse_args()
    args.early_stopping = int_to_bool(args.early_stopping)
    args.freeze_backbone = int_to_bool(args.freeze_backbone)
    args.enable_prefix_tuning = int_to_bool(args.enable_prefix_tuning)
    args.d_model = args.patch_length * args.d_model_scale
    args.decoder_d_model = args.patch_length * args.decoder_d_model_scale

    # Calculate number of gpus
    if args.num_gpus is None:
        args.num_gpus = torch.cuda.device_count()
        logger.info(f"Automatically calculated number of GPUs ={args.num_gpus}")

    # Create save directory
    args.save_dir = os.path.join(
        args.save_dir,
        f"TTM_cl-{args.context_length}_fl-{args.forecast_length}_pl-{args.patch_length}_apl-{args.adaptive_patching_levels}_ne-{args.num_epochs}_es-{args.early_stopping}",
    )
    os.makedirs(args.save_dir, exist_ok=True)

    return args


def int_to_bool(value):  # pragma: no cover
    if value == 0:
        return False
    elif value == 1:
        return True
    else:
        raise argparse.ArgumentTypeError("Boolean value expected (0 or 1)")



# Copyright contributors to the TSFM project
#

import copy
import json
import os
import tempfile
from datetime import timedelta

import numpy as np
import pandas as pd
import pytest
from fastapi import HTTPException
from tsfminference.inference import InferenceRuntime
from tsfminference.inference_payloads import (
    ForecastingInferenceInput,
    ForecastingMetadataInput,
    ForecastingParameters,
    PredictOutput,
)


SERIES_LENGTH = int(os.getenv("TSFM_PROFILE_SERIES_LENGTH", 512))
FORECAST_LENGTH = 96
MODEL_ID = "mytest-tsfm/ttm-r1"
NUM_TIMESERIES = int(os.getenv("TSFM_PROFILE_NUM_TIMESERIES", 2))


@pytest.fixture(scope="module")
def ts_data_base() -> pd.DataFrame:
    # Generate a date range
    length = SERIES_LENGTH
    date_range = pd.date_range(start="2023-10-01", periods=length, freq="h")

    timeseries = []
    for idx in range(NUM_TIMESERIES):
        timeseries.append(
            pd.DataFrame(
                {
                    "date": date_range,
                    "ID": str(idx),
                    "VAL": np.random.rand(length),
                }
            )
        )

    return pd.concat(timeseries, ignore_index=True)


@pytest.fixture(scope="module")
def forecasting_input_base() -> ForecastingInferenceInput:
    # df: pd.DataFrame = ts_data_base
    schema: ForecastingMetadataInput = ForecastingMetadataInput(
        timestamp_column="date", id_columns=["ID"], target_columns=["VAL"]
    )
    parameters: ForecastingParameters = ForecastingParameters()
    input: ForecastingInferenceInput = ForecastingInferenceInput(
        model_id=MODEL_ID,
        schema=schema,
        parameters=parameters,
        data={
            "date": [
                "2024-10-18T01:00:21+00:00",
            ],
            "ID1": [
                "I1",
            ],
            "VAL": [
                10.0,
            ],
        },  # this should get replaced in each test case anyway,
    )
    return input


def _basic_result_checks(results: PredictOutput, df: pd.DataFrame):
    # expected length
    assert len(results) == FORECAST_LENGTH * NUM_TIMESERIES
    # expected start time
    answer = results["date"].iloc[0] - df["date"].iloc[-1]
    assert (
        answer == timedelta(hours=1) if isinstance(answer, timedelta) else answer == timedelta(hours=1).total_seconds()
    )
    # expected end time
    answer = results["date"].iloc[-1] - df["date"].iloc[-1]
    assert (
        answer == timedelta(hours=FORECAST_LENGTH)
        if isinstance(answer, timedelta)
        else answer == timedelta(hours=FORECAST_LENGTH).total_seconds()
    )


def test_forecast_with_decimal_freq(ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput):
    input: ForecastingInferenceInput = forecasting_input_base
    input: ForecastingInferenceInput = copy.deepcopy(input)
    input.schema.freq = "3600.0s"  # 1-hr
    df = copy.deepcopy(ts_data_base)
    input.data = df.to_dict(orient="list")
    runtime: InferenceRuntime = InferenceRuntime()
    po: PredictOutput = runtime.forecast(input=input)
    results = pd.DataFrame.from_dict(po.results[0])
    _basic_result_checks(results, df)

    # without a period specifier
    input: ForecastingInferenceInput = forecasting_input_base
    input: ForecastingInferenceInput = copy.deepcopy(input)
    input.schema.freq = "3600.0"  # 1-hr
    df = copy.deepcopy(ts_data_base)
    df["date"] = df["date"].values.astype("int64") // 10**9
    input.data = df.to_dict(orient="list")
    runtime: InferenceRuntime = InferenceRuntime()
    po: PredictOutput = runtime.forecast(input=input)
    results = pd.DataFrame.from_dict(po.results[0])
    _basic_result_checks(results, df)


def test_forecast_with_single_character_column_name(
    ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput
):
    input: ForecastingInferenceInput = forecasting_input_base
    input: ForecastingInferenceInput = copy.deepcopy(input)
    input.schema.target_columns = ["V"]
    input.schema.timestamp_column = "D"
    input.schema.id_columns = ["I"]
    df = copy.deepcopy(ts_data_base)
    df = df.rename(columns={"VAL": "V"})
    df = df.rename(columns={"date": "D"})
    df = df.rename(columns={"ID": "I"})

    input.data = df.to_dict(orient="list")
    runtime: InferenceRuntime = InferenceRuntime()
    po: PredictOutput = runtime.forecast(input=input)
    results = pd.DataFrame.from_dict(po.results[0])

    # change column names back so _basic_result_checks works
    df = df.rename(columns={"V": "VAL"})
    df = df.rename(columns={"D": "date"})
    df = df.rename(columns={"I": "ID"})

    results = results.rename(columns={"V": "VAL"})
    results = results.rename(columns={"D": "date"})
    results = results.rename(columns={"I": "ID"})

    _basic_result_checks(results, df)


def test_forecast_with_good_data(ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput):
    input = forecasting_input_base
    model_id = input.model_id
    df = ts_data_base if int(os.environ.get("TSFM_TESTS_AS_PROFILER", "0")) == 0 else copy.deepcopy(ts_data_base)
    input.data = df.to_dict(orient="list")

    # useful for generating sample payload files
    if int(os.environ.get("TSFM_TESTS_DO_VERBOSE_DUMPS", "0")) == 1:
        with open(f"{tempfile.gettempdir()}/{model_id}.payload.json", "w") as out:
            foo = copy.deepcopy(df)
            foo["date"] = foo["date"].apply(lambda x: x.isoformat())
            json.dump(foo.to_dict(orient="list"), out)

    runtime: InferenceRuntime = InferenceRuntime()
    po: PredictOutput = runtime.forecast(input=input)
    results = pd.DataFrame.from_dict(po.results[0])
    _basic_result_checks(results, df)


def test_forecast_with_schema_missing_target_columns(
    ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput
):
    input = forecasting_input_base
    input.schema.target_columns = []
    df = copy.deepcopy(ts_data_base)
    input.data = df.to_dict(orient="list")
    runtime: InferenceRuntime = InferenceRuntime()
    po: PredictOutput = runtime.forecast(input=input)
    results = pd.DataFrame.from_dict(po.results[0])
    _basic_result_checks(results, df)


def test_forecast_with_integer_timestamps(
    ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput
):
    input: ForecastingInferenceInput = copy.deepcopy(forecasting_input_base)
    df = copy.deepcopy(ts_data_base)

    timestamp_column = input.schema.timestamp_column
    df[timestamp_column] = df[timestamp_column].astype(int)
    df[timestamp_column] = range(1, SERIES_LENGTH * NUM_TIMESERIES + 1)
    input.data = df.to_dict(orient="list")
    runtime: InferenceRuntime = InferenceRuntime()
    po: PredictOutput = runtime.forecast(input=input)
    results = pd.DataFrame.from_dict(po.results[0])
    assert results[timestamp_column].iloc[0] == SERIES_LENGTH + 1
    assert results[timestamp_column].iloc[-1] - df[timestamp_column].iloc[-1] == FORECAST_LENGTH
    assert results.dtypes[timestamp_column] == df.dtypes[timestamp_column]


def test_forecast_with_float_timestamps(ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput):
    input: ForecastingInferenceInput = copy.deepcopy(forecasting_input_base)
    df = copy.deepcopy(ts_data_base)

    timestamp_column = input.schema.timestamp_column
    df[timestamp_column] = [float(0.0) + x for x in range(1, SERIES_LENGTH * NUM_TIMESERIES + 1)]
    input.data = df.to_dict(orient="list")
    runtime: InferenceRuntime = InferenceRuntime()
    po: PredictOutput = runtime.forecast(input=input)
    results = pd.DataFrame.from_dict(po.results[0])
    assert results[timestamp_column].iloc[0] == SERIES_LENGTH + 1
    assert results[timestamp_column].iloc[-1] - df[timestamp_column].iloc[-1] == FORECAST_LENGTH
    assert results.dtypes[timestamp_column] == df.dtypes[timestamp_column]


def test_forecast_with_bogus_timestamps(ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput):
    input: ForecastingInferenceInput = copy.deepcopy(forecasting_input_base)
    df = copy.deepcopy(ts_data_base)

    timestamp_column = input.schema.timestamp_column
    df[timestamp_column] = df[timestamp_column].astype(str)
    df[timestamp_column] = [str(x) for x in range(1, SERIES_LENGTH * NUM_TIMESERIES + 1)]
    input.data = df.to_dict(orient="list")
    runtime: InferenceRuntime = InferenceRuntime()
    with pytest.raises(HTTPException) as _:
        runtime.forecast(input=input)


def test_forecast_with_bogus_values(ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput):
    input: ForecastingInferenceInput = copy.deepcopy(forecasting_input_base)
    df = copy.deepcopy(ts_data_base)
    df["VAL"] = df["VAL"].astype(str)
    df["VAL"] = [str(x) for x in range(1, SERIES_LENGTH * NUM_TIMESERIES + 1)]
    input.data = df.to_dict(orient="list")
    runtime: InferenceRuntime = InferenceRuntime()
    with pytest.raises(HTTPException) as _:
        runtime.forecast(input=input)


def test_forecast_with_bogus_model_id(ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput):
    input: ForecastingInferenceInput = copy.deepcopy(forecasting_input_base)
    df = copy.deepcopy(ts_data_base)
    input.data = df.to_dict(orient="list")
    input.model_id = "hoo-hah"

    runtime: InferenceRuntime = InferenceRuntime()
    with pytest.raises(HTTPException) as _:
        runtime.forecast(input=input)


def test_forecast_with_insufficient_context_length(
    ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput
):
    input: ForecastingInferenceInput = copy.deepcopy(forecasting_input_base)
    df = copy.deepcopy(ts_data_base)
    df = df.iloc[0:-100]

    input.data = df.to_dict(orient="list")

    runtime: InferenceRuntime = InferenceRuntime()
    with pytest.raises(HTTPException) as _:
        runtime.forecast(input=input)


@pytest.mark.skip
def test_forecast_with_nan_data(ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput):
    input: ForecastingInferenceInput = copy.deepcopy(forecasting_input_base)
    df = copy.deepcopy(ts_data_base)
    df.iloc[0, df.columns.get_loc("VAL")] = np.nan

    input.data = df.to_dict(orient="list")

    runtime: InferenceRuntime = InferenceRuntime()
    # with pytest.raises(HTTPException) as _:
    runtime.forecast(input=input)


# @pytest.mark.skip
def test_forecast_with_missing_row(ts_data_base: pd.DataFrame, forecasting_input_base: ForecastingInferenceInput):
    input: ForecastingInferenceInput = copy.deepcopy(forecasting_input_base)
    df = copy.deepcopy(ts_data_base)
    df = df.drop(index=10)

    # append a row to give it the correct length
    # don't forget to update the timestamp accordingly in the
    # appended row

    input.data = df.to_dict(orient="list")

    runtime: InferenceRuntime = InferenceRuntime()
    with pytest.raises(HTTPException) as _:
        runtime.forecast(input=input)



import csv
import json
import math
import os
import tempfile
import warnings

import numpy as np
import torch
import transformers
from packaging import version
from sklearn.model_selection import StratifiedKFold
from torch.optim import AdamW
from torch.optim.lr_scheduler import OneCycleLR
from torch.utils.data import DataLoader, Subset, random_split
from transformers import EarlyStoppingCallback, Trainer, TrainingArguments, set_seed
from transformers.data.data_collator import default_data_collator
from transformers.trainer_utils import RemoveColumnsCollator

from tsfm_public.models.tspulse import TSPulseForClassification
from tsfm_public.toolkit.dataset import ClassificationDFDataset
from tsfm_public.toolkit.lr_finder import optimal_lr_finder
from tsfm_public.toolkit.time_series_classification_preprocessor import TimeSeriesClassificationPreprocessor
from tsfm_public.toolkit.util import convert_tsfile_to_dataframe


warnings.filterwarnings("ignore")

REQUIRED_TORCH_VERSION = "2.4.0+cu121"
REQUIRED_TRANSFORMERS_VERSION = "4.44.0"

if torch.__version__ != REQUIRED_TORCH_VERSION:
    print("torch.__version__ : ", torch.__version__)
    raise RuntimeError(
        f"Torch version mismatch: Please use {REQUIRED_TORCH_VERSION} for reproducibility of classification scores."
    )

MIN_TRANSFORMERS_VERSION = "4.44.0"
MAX_TRANSFORMERS_VERSION = "4.50.3"

current_version = transformers.__version__

if not (
    version.parse(MIN_TRANSFORMERS_VERSION)
    <= version.parse(current_version)
    <= version.parse(MAX_TRANSFORMERS_VERSION)
):
    raise RuntimeError(
        f"Transformers version mismatch: Please use a version between {MIN_TRANSFORMERS_VERSION} and {MAX_TRANSFORMERS_VERSION} "
        f"for reproducibility of classification scores. Found version: {current_version}"
    )

OUT_DIR = "tspulse_finetuned_models/"

json_file = "config.json"
with open(json_file, "r") as file:
    clf_params = json.load(file)


def main(dataset_name):
    seed = 42
    set_seed(seed)
    path = f"datasets/{dataset_name}/{dataset_name}_TRAIN.ts"  # train

    df_base = convert_tsfile_to_dataframe(
        path,
        return_separate_X_and_y=False,
    )
    label_column = "class_vals"
    input_columns = [f"dim_{i}" for i in range(df_base.shape[1] - 1)]

    tsp = TimeSeriesClassificationPreprocessor(
        input_columns=input_columns,
        label_column=label_column,
        scaling=True,
    )

    tsp.train(df_base)

    df_prep = tsp.preprocess(df_base)
    base_dataset = ClassificationDFDataset(
        df_prep,
        id_columns=[],
        timestamp_column=None,
        input_columns=input_columns,
        label_column=label_column,
        context_length=512,
        static_categorical_columns=[],
        stride=1,
        enable_padding=False,
        full_series=True,
    )

    path = f"datasets/{dataset_name}/{dataset_name}_TEST.ts"  # test

    df_test = convert_tsfile_to_dataframe(
        path,
        return_separate_X_and_y=False,
    )
    label_column = "class_vals"
    input_columns = [f"dim_{i}" for i in range(df_test.shape[1] - 1)]

    tsp = TimeSeriesClassificationPreprocessor(
        input_columns=input_columns,
        label_column=label_column,
        scaling=True,
    )

    tsp.train(df_test)

    df_prep = tsp.preprocess(df_test)

    test_dataset = ClassificationDFDataset(
        df_prep,
        id_columns=[],
        timestamp_column=None,
        input_columns=input_columns,
        label_column=label_column,
        context_length=512,
        static_categorical_columns=[],
        stride=1,
        enable_padding=False,
        full_series=True,
    )

    # validation set
    if dataset_name == "DuckDuckGeese":
        data_dict = clf_params[dataset_name]["DATA_PARAMS"]

        def k_fold_cv(skf, dataset, kth_fold):
            y = []
            for idx in range(len(dataset)):
                y.append(dataset[idx]["target_values"])

            for i, (train_indices, val_indices) in enumerate(skf.split(np.zeros(len(y)), y)):
                if i == kth_fold:
                    train_dataset = Subset(dataset, train_indices)
                    val_dataset = Subset(dataset, val_indices)
            return train_dataset, val_dataset

        skf = StratifiedKFold(n_splits=data_dict["num_folds"], shuffle=True)
        train_dataset, valid_dataset = k_fold_cv(skf, base_dataset, data_dict["kth_fold"])
    else:
        dataset_size = len(base_dataset)
        print(dataset_size)
        split_valid_ratio = 0.1
        val_size = int(split_valid_ratio * dataset_size)  # 10% valid split
        train_size = dataset_size - val_size
        train_dataset, valid_dataset = random_split(base_dataset, [train_size, val_size])

    config_dict = clf_params[dataset_name]["MODEL_PARAMS"]
    config_dict["loss"] = "cross_entropy"
    config_dict["ignore_mismatched_sizes"] = True

    config_dict["num_input_channels"] = tsp.num_input_channels
    config_dict["num_targets"] = df_base["class_vals"].nunique()

    model = TSPulseForClassification.from_pretrained(
        "ibm-granite/granite-timeseries-tspulse-r1", revision="tspulse-block-dualhead-512-p16-r1", **config_dict
    )
    model = model.to("cuda").float()

    # Freezing Backbone except patch embedding layer....
    for param in model.backbone.parameters():
        param.requires_grad = False

    for param in model.backbone.time_encoding.parameters():
        param.requires_grad = True
    for param in model.backbone.fft_encoding.parameters():
        param.requires_grad = True

    temp_dir = tempfile.mkdtemp()

    suggested_lr = None

    train_dict = clf_params[dataset_name]["TRAINING_PARAMS"]
    EPOCHS = train_dict["num_train_epochs"]
    BATCH_SIZE = train_dict["per_device_train_batch_size"]
    eval_accumulation_steps = train_dict["eval_accumulation_steps"]
    NUM_WORKERS = 1
    NUM_GPUS = 1

    set_seed(42)
    if suggested_lr is None:
        lr, model = optimal_lr_finder(
            model,
            train_dataset,
            batch_size=BATCH_SIZE,
        )
        suggested_lr = lr
    print("Suggested LR : ", suggested_lr)
    finetune_args = TrainingArguments(
        output_dir=temp_dir,
        overwrite_output_dir=True,
        learning_rate=suggested_lr,
        num_train_epochs=EPOCHS,
        do_eval=True,
        eval_strategy="epoch",
        per_device_train_batch_size=BATCH_SIZE,
        per_device_eval_batch_size=BATCH_SIZE,
        eval_accumulation_steps=eval_accumulation_steps,
        dataloader_num_workers=NUM_WORKERS,
        report_to="tensorboard",
        save_strategy="epoch",
        logging_strategy="epoch",
        save_total_limit=1,
        logging_dir=os.path.join(OUT_DIR, "output"),  # Make sure to specify a logging directory
        load_best_model_at_end=True,  # Load the best model when training ends
        metric_for_best_model="eval_loss",  # Metric to monitor for early stopping
        greater_is_better=False,  # For loss
    )

    # Create the early stopping callback
    early_stopping_callback = EarlyStoppingCallback(
        early_stopping_patience=clf_params[dataset_name]["EARYL_STOPPING_PARAMS"][
            "early_stopping_patience"
        ],  # Number of epochs with no improvement after which to stop
        early_stopping_threshold=0.0001,  # Minimum improvement required to consider as improvement
    )

    # Optimizer and scheduler
    optimizer = AdamW(model.parameters(), lr=suggested_lr)
    scheduler = OneCycleLR(
        optimizer,
        suggested_lr,
        epochs=EPOCHS,
        steps_per_epoch=math.ceil(len(train_dataset) / (BATCH_SIZE * NUM_GPUS)),
    )

    finetune_trainer = Trainer(
        model=model,
        args=finetune_args,
        train_dataset=train_dataset,
        eval_dataset=valid_dataset,
        callbacks=[early_stopping_callback],
        optimizers=(optimizer, scheduler),
    )

    # Fine tune
    finetune_trainer.train()

    predictions_dict = finetune_trainer.predict(test_dataset)
    preds_np = predictions_dict.predictions[0]

    remove_columns_collator = RemoveColumnsCollator(
        data_collator=default_data_collator,
        signature_columns=["target_values"],
        logger=None,
        description=None,
        model_name="temp",
    )

    test_dataloader = DataLoader(test_dataset, batch_size=32, shuffle=False, collate_fn=remove_columns_collator)
    target_list = []
    for batch in test_dataloader:
        batch_labels = batch["target_values"].numpy()
        target_list.append(batch_labels)
    targets_np = np.concatenate(target_list, axis=0)
    test_accuracy = np.mean(targets_np == np.argmax(preds_np, axis=1))
    print("test_accuracy : ", test_accuracy)

    output_file = "tspulse_uea_classification_accuracies.csv"

    if not os.path.exists(output_file):
        with open(output_file, mode="w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(["Dataset", "Accuracy"])

    with open(output_file, mode="a", newline="") as file:
        writer = csv.writer(file)
        writer.writerow([dataset_name, test_accuracy])


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--datasets", nargs="+", required=True, help="List of UEA dataset names")
    args = parser.parse_args()

    print("Datasets : ", args.datasets)

    for dataset_name in args.datasets:
        main(dataset_name)



# Copyright contributors to the TSFM project
#

"""Tests the time series preprocessor and functions"""

import numpy as np
import pandas as pd
import pytest

from tsfm_public.models.tinytimemixer import TinyTimeMixerConfig, TinyTimeMixerForPrediction
from tsfm_public.models.tspulse import TSPulseConfig, TSPulseForReconstruction
from tsfm_public.toolkit.ad_helpers import AnomalyScoreMethods
from tsfm_public.toolkit.conformal import (
    NonconformityScores,
    PostHocProbabilisticMethod,
    PostHocProbabilisticProcessor,
)
from tsfm_public.toolkit.time_series_anomaly_detection_pipeline import (
    TimeSeriesAnomalyDetectionPipeline,
)
from tsfm_public.toolkit.time_series_forecasting_pipeline import TimeSeriesForecastingPipeline


@pytest.fixture(scope="module")
def example_dataset():
    n_variables: int = 2
    target_variables = [f"X{i + 1}" for i in range(n_variables)]
    data = np.array(
        [np.convolve(np.random.normal(0, 1, 1000), np.ones(15) / 15, "same") for _ in range(n_variables)]
    ).T
    timestamp = pd.date_range("2021-01-01", periods=len(data), freq=pd.Timedelta(5, "minute"))
    df = pd.DataFrame(data, columns=target_variables)
    df["timestamp"] = timestamp
    return target_variables, df


@pytest.mark.parametrize(
    "method",
    [
        [AnomalyScoreMethods.TIME_RECONSTRUCTION.value],
        [AnomalyScoreMethods.FREQUENCY_RECONSTRUCTION.value],
        [AnomalyScoreMethods.PREDICTIVE.value],
        [
            AnomalyScoreMethods.TIME_RECONSTRUCTION.value,
            AnomalyScoreMethods.FREQUENCY_RECONSTRUCTION.value,
            AnomalyScoreMethods.PREDICTIVE.value,
        ],
    ],
)
def test_tsad_tspulse_pipeline_defaults(example_dataset, method):
    target_variables, dataset = example_dataset
    params = {}
    params.update(
        context_length=512,
        patch_length=8,
        mask_block_length=2,
        num_input_channels=len(target_variables),
        patch_stride=8,
        mode="common_channel",
        gated_attn=True,
        use_positional_encoding=False,
        self_attn=False,
        head_aggregation="max_pool",
        fuse_fft=True,
        use_learnable_mask_token=True,
        prediction_length=4,
        fft_time_add_forecasting_pt_loss=True,
        channel_mix_init="identity",
    )

    model = TSPulseForReconstruction(TSPulseConfig(**params))

    tspipe = TimeSeriesAnomalyDetectionPipeline(
        model,
        prediction_mode=method,
        timestamp_column="timestamp",
        target_columns=target_variables,
        aggregation_length=32,
    )

    assert tspipe._preprocess_params["prediction_length"] == 1
    assert tspipe._preprocess_params["context_length"] == model.config.context_length

    result = tspipe(dataset)
    assert result.shape[0] == dataset.shape[0]
    assert "anomaly_score" in result

    result = tspipe(dataset, expand_score=True)
    assert result.shape[0] == dataset.shape[0]
    for tgt in target_variables:
        assert f"{tgt}_anomaly_score" in result

    result = tspipe(dataset, report_mode=True)
    assert result.shape[0] == dataset.shape[0]
    assert "selected_mode" in result

    result = tspipe(dataset, expand_score=True, report_mode=True)
    assert result.shape[0] == dataset.shape[0]
    for tgt in target_variables:
        assert f"{tgt}_selected_mode" in result


@pytest.mark.parametrize(
    "method",
    [
        [AnomalyScoreMethods.PREDICTIVE.value],
        [AnomalyScoreMethods.MEAN_DEVIATION.value],
        [AnomalyScoreMethods.MEAN_DEVIATION.value, AnomalyScoreMethods.PREDICTIVE.value],
    ],
)
def test_tsad_tinytimemixer_pipeline_defaults(example_dataset, method):
    target_variables, dataset = example_dataset
    model = TinyTimeMixerForPrediction(
        TinyTimeMixerConfig(context_length=120, prediction_length=60, num_input_channels=len(target_variables))
    )

    tspipe = TimeSeriesAnomalyDetectionPipeline(
        model,
        prediction_mode=method,
        timestamp_column="timestamp",
        target_columns=["X1", "X2"],
    )
    assert tspipe._preprocess_params["prediction_length"] == model.config.prediction_length
    assert tspipe._preprocess_params["context_length"] == model.config.context_length
    result = tspipe(dataset)
    assert result.shape[0] == dataset.shape[0]
    assert "anomaly_score" in result

    result = tspipe(dataset, expand_score=True)
    assert result.shape[0] == dataset.shape[0]
    for tgt in target_variables:
        assert f"{tgt}_anomaly_score" in result


@pytest.mark.parametrize(
    "method",
    [
        [AnomalyScoreMethods.PROBABILISTIC.value],
        [AnomalyScoreMethods.PROBABILISTIC.value, AnomalyScoreMethods.MEAN_DEVIATION.value],
    ],
)
def test_tsad_tinytimemixer_pipeline_probabilistic(example_dataset, method):
    target_variables, dataset = example_dataset
    model = TinyTimeMixerForPrediction(
        TinyTimeMixerConfig(context_length=120, prediction_length=60, num_input_channels=len(target_variables))
    )

    prob_proc = PostHocProbabilisticProcessor(
        window_size=200,
        quantiles=[0.25, 0.75],
        nonconformity_score=NonconformityScores.ABSOLUTE_ERROR.value,
        method=PostHocProbabilisticMethod.CONFORMAL.value,
        smoothing_length=None,
    )

    fpipe = TimeSeriesForecastingPipeline(
        model, timestamp_column="timestamp", id_columns=[], target_columns=target_variables, device="cpu"
    )
    forecasts = fpipe(dataset)

    prediction_length = 60
    prediction_columns = [f"{c}_prediction" for c in target_variables]
    ground_truth_columns = target_variables
    y = forecasts[prediction_columns].values
    predictions = np.array([np.stack(z) for z in y]).transpose(0, 2, 1)
    predictions = predictions[:-prediction_length, ...]
    x = forecasts[ground_truth_columns].values
    ground_truth = np.array([np.stack(z) for z in x]).transpose(0, 2, 1)
    ground_truth = ground_truth[:-prediction_length, ...]

    prob_proc.train(y_cal_gt=ground_truth, y_cal_pred=predictions)

    tspipe = TimeSeriesAnomalyDetectionPipeline(
        model,
        prediction_mode=method,
        probabilistic_processor=prob_proc,
        timestamp_column="timestamp",
        target_columns=["X1", "X2"],
        device="cpu",
    )
    assert tspipe._preprocess_params["prediction_length"] == model.config.prediction_length
    assert tspipe._preprocess_params["context_length"] == model.config.context_length
    result = tspipe(dataset)
    assert result.shape[0] == dataset.shape[0]
    assert "anomaly_score" in result

    result = tspipe(dataset, expand_score=True)
    assert result.shape[0] == dataset.shape[0]
    for tgt in target_variables:
        assert f"{tgt}_anomaly_score" in result

    result = tspipe(
        dataset,
        prediction_mode=f"{AnomalyScoreMethods.PROBABILISTIC.value}+{AnomalyScoreMethods.MEAN_DEVIATION.value}",
        expand_score=True,
    )
    assert result.shape[0] == dataset.shape[0]
    for tgt in target_variables:
        assert f"{tgt}_anomaly_score" in result



import pathlib
import subprocess
import sys
import tempfile
from subprocess import CompletedProcess


def test_ftmain_good_input():
    with tempfile.TemporaryDirectory() as target_dir:
        model_name = "afinetuned_model"
        args = [
            "python",
            "tsfmfinetuning/ftmain.py",
            "--payload",
            "./data/ftpayload.json",
            "--target_dir",
            target_dir,
            "--model_name",
            model_name,
            "--config_file",
            "tsfmfinetuning/default_config.yml",
            "--model_arch",
            "ttm",
            "--task_type",
            "forecasting",
        ]

        other_popen_kwargs = {}
        cp: CompletedProcess = subprocess.run(
            args,
            stdin=None,
            input=None,
            capture_output=True,
            text=True,
            shell=False,
            cwd=None,
            timeout=None,
            check=False,
            encoding=None,
            errors=None,
            env=None,
            universal_newlines=None,
            **other_popen_kwargs,
        )

        print(cp.stdout)
        print(cp.stderr)
        sys.stdout.flush()
        assert cp.returncode == 0
        tdir = pathlib.Path(target_dir) / model_name
        assert (tdir / "config.json").exists() and (tdir / "config.json").stat().st_size > 0

        # make sure our tracking log exists
        logfile = pathlib.Path(target_dir) / "output" / "training_logs.jsonl"
        assert logfile.exists() and logfile.stat().st_size > 0



import os
from collections import defaultdict

import numpy as np
from sklearn.preprocessing import StandardScaler

from tsfm_public.toolkit.dataset import ClassificationDFDataset
from tsfm_public.toolkit.time_series_classification_preprocessor import (
    TimeSeriesClassificationPreprocessor,
)
from tsfm_public.toolkit.util import convert_tsfile_to_dataframe


np.random.seed(0)


DATA_DIR = "./Univariate_ts"
DEFAULT_LIST_UCR_DSETS = [
    "ACSF1",
    "Adiac",
    "AllGestureWiimoteX",
    "AllGestureWiimoteY",
    "AllGestureWiimoteZ",
    "ArrowHead",
    "BME",
    "Beef",
    "BeetleFly",
    "BirdChicken",
    "CBF",
    "Car",
    "Chinatown",
    "ChlorineConcentration",
    "CinCECGTorso",
    "Coffee",
    "Computers",
    "CricketX",
    "CricketY",
    "CricketZ",
    "Crop",
    "DiatomSizeReduction",
    "DistalPhalanxOutlineAgeGroup",
    "DistalPhalanxOutlineCorrect",
    "DistalPhalanxTW",
    "DodgerLoopDay",
    "DodgerLoopGame",
    "DodgerLoopWeekend",
    "ECG200",
    "ECG5000",
    "ECGFiveDays",
    "EOGHorizontalSignal",
    "EOGVerticalSignal",
    "Earthquakes",
    "ElectricDevices",
    "EthanolLevel",
    "FaceAll",
    "FaceFour",
    "FacesUCR",
    "FiftyWords",
    "Fish",
    "FordA",
    "FordB",
    "FreezerRegularTrain",
    "FreezerSmallTrain",
    "Fungi",
    "GestureMidAirD1",
    "GestureMidAirD2",
    "GestureMidAirD3",
    "GesturePebbleZ1",
    "GesturePebbleZ2",
    "GunPoint",
    "GunPointAgeSpan",
    "GunPointMaleVersusFemale",
    "GunPointOldVersusYoung",
    "Ham",
    "HandOutlines",
    "Haptics",
    "Herring",
    "HouseTwenty",
    "InlineSkate",
    "InsectEPGRegularTrain",
    "InsectEPGSmallTrain",
    "InsectWingbeatSound",
    "ItalyPowerDemand",
    "LargeKitchenAppliances",
    "Lightning2",
    "Lightning7",
    "Mallat",
    "Meat",
    "MedicalImages",
    "MelbournePedestrian",
    "MiddlePhalanxOutlineAgeGroup",
    "MiddlePhalanxOutlineCorrect",
    "MiddlePhalanxTW",
    "MixedShapesRegularTrain",
    "MixedShapesSmallTrain",
    "MoteStrain",
    "NonInvasiveFetalECGThorax1",
    "NonInvasiveFetalECGThorax2",
    "OSULeaf",
    "OliveOil",
    "PLAID",
    "PhalangesOutlinesCorrect",
    "Phoneme",
    "PickupGestureWiimoteZ",
    "PigAirwayPressure",
    "PigArtPressure",
    "PigCVP",
    "Plane",
    "PowerCons",
    "ProximalPhalanxOutlineAgeGroup",
    "ProximalPhalanxOutlineCorrect",
    "ProximalPhalanxTW",
    "RefrigerationDevices",
    "Rock",
    "ScreenType",
    "SemgHandGenderCh2",
    "SemgHandMovementCh2",
    "SemgHandSubjectCh2",
    "ShakeGestureWiimoteZ",
    "ShapeletSim",
    "ShapesAll",
    "SmallKitchenAppliances",
    "SmoothSubspace",
    "SonyAIBORobotSurface1",
    "SonyAIBORobotSurface2",
    "StarLightCurves",
    "Strawberry",
    "SwedishLeaf",
    "Symbols",
    "SyntheticControl",
    "ToeSegmentation1",
    "ToeSegmentation2",
    "Trace",
    "TwoLeadECG",
    "TwoPatterns",
    "UMD",
    "UWaveGestureLibraryAll",
    "UWaveGestureLibraryX",
    "UWaveGestureLibraryY",
    "UWaveGestureLibraryZ",
    "Wafer",
    "Wine",
    "WordSynonyms",
    "Worms",
    "WormsTwoClass",
    "Yoga",
]


LABEL_COLUMN = "class_vals"
INPUT_COLUMNS = ["dim_0"]  # because univariate ts
CONTEXT_LENGTH = 512


min_num_req_samples = 10
mixed_ts = []
mixed_dset_names = []
mixed_dset_classes = []

# pick datasets with more than min_num_req_samples
# then, perform sampling to create a balanced data
for dataname in DEFAULT_LIST_UCR_DSETS:
    skip_this_dataset = False
    print(dataname)

    data_path = os.path.join(DATA_DIR, dataname, f"{dataname}_TRAIN.ts")
    df_train = convert_tsfile_to_dataframe(data_path, return_separate_X_and_y=False)
    tsp = TimeSeriesClassificationPreprocessor(
        input_columns=INPUT_COLUMNS,
        label_column=LABEL_COLUMN,
    )

    tsp.train(df_train)
    df_prep = tsp.preprocess(df_train)
    train_dataset = ClassificationDFDataset(
        df_prep,
        id_columns=[],
        timestamp_column=None,
        input_columns=INPUT_COLUMNS,
        label_column=LABEL_COLUMN,
        context_length=CONTEXT_LENGTH,
        static_categorical_columns=[],
        enable_padding=False,
        full_series=True,
    )
    print(f"Total Train samples: {len(train_dataset)}")

    label_counts = defaultdict(int)
    for sample in train_dataset:
        label = sample["target_values"]
        label_counts[label.item()] += 1

    label_counts = dict(label_counts)
    sorted_label_counts = dict(sorted(label_counts.items()))
    print(dataname, " ", sorted_label_counts)

    for k, v in sorted_label_counts.items():
        if v < min_num_req_samples:
            print(
                f"[NOTE] : {dataname} have less than the required num. of samples {min_num_req_samples} in class {k}, so skipping this dataset"
            )
            skip_this_dataset = True
            break

    if skip_this_dataset:
        continue

    class_indices = defaultdict(list)
    for idx in range(len(train_dataset)):
        label = int(train_dataset[idx]["target_values"])
        # label = int(label)
        class_indices[label].append(idx)

    for class_label, indices in class_indices.items():
        sub_indices = np.random.choice(indices, size=10, replace=False)
        for index in sub_indices:
            datapoint = train_dataset[index]
            datapoint_ts = datapoint["past_values"]  # [T, C]
            datapoint_ts = StandardScaler().fit_transform(datapoint_ts)

            mixed_ts.append(datapoint_ts)
            mixed_dset_names.append(dataname)
            mixed_dset_classes.append(class_label)

mixed_ts, mixed_dset_classes, mixed_dset_names = (
    np.array(mixed_ts),
    np.array(mixed_dset_classes),
    np.array(mixed_dset_names),
)
print(
    "total samples : ",
    mixed_ts.shape,
    mixed_dset_classes.shape,
    mixed_dset_names.shape,
)
np.savez("ucr_for_search.npz", ts=mixed_ts, names=mixed_dset_names, classes=mixed_dset_classes)
